#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for Device Control."""

from __future__ import with_statement

import os
import re
from iobase.xmldata import XmlNode
from iobase.cmdexec import (CmdExec, PIPE, NULL)
from iobase.exception import InvalidArgError
from pal_info import (ReqTypes, Results)
import libpal.palcmd as palcmd
import libpal.util as util
from pal_config import (
    UDEVADM_CMD, DEV_DIR,
    DISK_X_MODEL_FILE, DISK_X_REV_FILE, DISK_X_VENDOR_FILE, DISK_X_SIZE_FILE,
    SATA_HDD_DEV_LIST, USB_HDD_DEV_LIST,
    USB_LP_DEV_LIST,
    USB_UPS_DEV_LIST, REAL_USB_UPS_LIST,
    SERIAL_UPS_DEV_LIST, REAL_SERIAL_UPS_LIST,
    UPS_DEV_LIST,
)

### XML Path ###
NAME_PATH = "./device/name"

### XML Tag Name ###
DEVCTRL_NAME = "devctrl"
DEVICE_NAME = "device"
DEVICELIST_NAME = "devicelist"
DISK_NAME = "disk"
USB_NAME = "usb"
MODEL_NAME = "model"
REV_NAME = "rev"
VENDOR_NAME = "vendor"
SIZE_NAME = "size"
NAME_NAME = "name"
PID_NAME = "pid"
VID_NAME = "vid"
MANUFACTURER_NAME = "manufacturer"
PRODUCT_NAME = "product"
SERIAL_NAME = "serial"

### Command and argment ###
VAR = "%s"
UDEV_SETTLE_ARG = [UDEVADM_CMD, "settle"]

# FORMAT:  udevadm --attribute-walk --name /dev/[devicename]
UDEV_ATTRIBUTE_ARG_FORMAT = [UDEVADM_CMD, "info", "--attribute-walk",
                             "--name",  VAR]
UDEV_ATTRIBUTE_VAR_INDEX_LIST = [4]

### DEVICE TYPE ###
SATA_HDD_TYPE = "SATAD"
USB_HDD_TYPE = "USBD"
USB_LPR_TYPE = "USBLPR"
USB_UPS_TYPE = " USBUPS"
SERIAL_UPS_TYPE = "SERUPS"
UNKNOWN_TYPE = "UNKNOWN"

ALL_REAL_DEV_LIST = SATA_HDD_DEV_LIST + USB_HDD_DEV_LIST + \
                    REAL_USB_UPS_LIST + REAL_SERIAL_UPS_LIST + \
                    USB_LP_DEV_LIST


class PalDevice(palcmd.PalCmd):
    """Pal Device implementation class."""

    def __validate_root_tag_name(self, entry):
        """Validate the root tag name."""

        root_tag_name = entry.get().get_node_name()
        if DEVCTRL_NAME != root_tag_name:
            raise InvalidArgError("invalid root tag..")

    def __convert_real_devname(self, dev_name):
        """Convert synbol device name to real device name.
           If dev_name is already real this method just returns it.
        """

        dev_path = os.path.join(DEV_DIR, dev_name)
        return os.path.basename(os.path.realpath(dev_path))

    def __get_device_type(self, dev_name):
        """Get device type of specified device"""

        dev_type_list_pair = \
            {SATA_HDD_TYPE: SATA_HDD_DEV_LIST,
             USB_HDD_TYPE: USB_HDD_DEV_LIST,
             USB_LPR_TYPE: USB_LP_DEV_LIST,
             USB_UPS_TYPE: USB_UPS_DEV_LIST + REAL_USB_UPS_LIST,
             SERIAL_UPS_TYPE: SERIAL_UPS_DEV_LIST}

        for dev_type, dev_list in dev_type_list_pair.items():
            if dev_name in dev_list:
                return dev_type

        return UNKNOWN_TYPE

    def __force_make_dirs(self, dir_path):
        """make specified directory."""

        if not os.path.isdir(dir_path):
            os.makedirs(dir_path)

    def __execute_cmd(self, is_enable_exception, is_wait,
                      std_out, std_err,
                      cmd_fmt, cmd_var_idx_list, variable_list):
        """Execute command and return Popen object
              cmd_fmt - specify command format list
              cmd_var_idx_list - specify variable index of cmd_fmt
              variable_list - variables corresponding to cmd_var_idx_list
        """

        cmd = cmd_fmt[:]
        for (loop_idx, var_idx) in enumerate(cmd_var_idx_list):
            cmd[var_idx] = variable_list[loop_idx]

        proc = CmdExec(cmd, stdout=std_out, stderr=std_err,
                       raise_exc=is_enable_exception).execute()

        if is_wait:
            proc.wait()

        return proc

    def __load_value(self, file_name):

        with open(file_name) as fh:
            return fh.read().rstrip()

    def __get_disk_info(self, disk_name):
        """Get specified disk info."""

        key_file_pair = {MODEL_NAME: DISK_X_MODEL_FILE,
                         REV_NAME: DISK_X_REV_FILE,
                         VENDOR_NAME: DISK_X_VENDOR_FILE,
                         SIZE_NAME: DISK_X_SIZE_FILE}

        real_disk_name = self.__convert_real_devname(disk_name)

        info_dict = {}
        for key, file_name in key_file_pair.items():
            info_dict[key] = self.__load_value(file_name % real_disk_name)
        return info_dict

    def __get_dev_info_as_usb(self, dev_name):
        """Get specified USB device info and update tmp file."""

        is_usb = False
        idProduct = ""
        idVendor = ""
        manufacturer = ""
        product = ""
        serial = ""
        version = ""

        proc = self.__execute_cmd(True, True,
                                  PIPE, NULL,
                                  UDEV_ATTRIBUTE_ARG_FORMAT,
                                  UDEV_ATTRIBUTE_VAR_INDEX_LIST,
                                  [os.path.join(DEV_DIR, dev_name)])
        info_dict = {}

        for line in proc.stdout:
            line = line.rstrip()

            if re.search("DRIVERS{0,1}==\"usb\"", line):
                if is_usb:
                    break

                is_usb = True
                continue

            if not idProduct:
                match = re.search("ATTRS{0,1}\{idProduct\}==\"(.*)\"", line)
                if match:
                    idProduct = match.group(1)

            if not idVendor:
                match = re.search("ATTRS{0,1}\{idVendor\}==\"(.*)\"", line)
                if match:
                    idVendor = match.group(1)

            if not manufacturer:
                match = re.search("ATTRS{0,1}\{manufacturer\}==\"(.*)\"", line)
                if match:
                    manufacturer = match.group(1)

            if not product:
                match = re.search("ATTRS{0,1}\{product\}==\"(.*)\"", line)
                if match:
                    product = match.group(1)

            if not serial:
                match = re.search("ATTRS{0,1}\{serial\}==\"(.*)\"", line)
                if match:
                    serial = match.group(1)

            if not version:
                match = re.search("ATTRS{0,1}\{version\}==\"\s*(\S+)\"", line)
                if match:
                    version = match.group(1)

            if is_usb and idProduct and idVendor and manufacturer \
                    and product and serial and version:
                break

        info_dict.update({PID_NAME: idProduct})
        info_dict.update({VID_NAME: idVendor})
        info_dict.update({MANUFACTURER_NAME: manufacturer})
        info_dict.update({PRODUCT_NAME: product})
        info_dict.update({SERIAL_NAME: serial})
        info_dict.update({REV_NAME: version})
        return info_dict

    def __append_disk_info(self, node, dev_name):
        """Get device info from tmp file and append specified node."""

        node.append_child(DISK_NAME)
        disk_node = node.get_node(DISK_NAME)

        info_dict = self.__get_disk_info(dev_name)
        tag_name_list = [MODEL_NAME, REV_NAME, VENDOR_NAME, SIZE_NAME]

        for tag_name in tag_name_list:
            disk_node.append_child(tag_name, info_dict[tag_name])
        return

    def __append_usb_info(self, node, dev_name):
        """Append usb device info into specified node"""

        node.append_child(USB_NAME)
        usb_node = node.get_node(USB_NAME)

        info_dict = self.__get_dev_info_as_usb(dev_name)
        tag_name_list = [PID_NAME, VID_NAME, MANUFACTURER_NAME,
                         PRODUCT_NAME, SERIAL_NAME, REV_NAME]

        for tag_name in tag_name_list:
            usb_node.append_child(tag_name, info_dict[tag_name])
        return

    def __append_sata_hdd_info(self, node, dev_name):

        self.__append_disk_info(node, dev_name)

        disk_path = os.path.realpath(os.path.join(DEV_DIR, dev_name))
        info_dict = util.get_sata_disk_info(disk_path)

        disk_node = node.get_node(DISK_NAME)
        disk_node.get_node(VENDOR_NAME).set_text(info_dict["vendor"])
        return

    def __append_usb_hdd_info(self, node, dev_name):

        self.__append_disk_info(node, dev_name)
        self.__append_usb_info(node, dev_name)
        return

    def __append_usb_lpr_info(self, node, dev_name):
        """Append usb printer info to specified node"""

        self.__append_usb_info(node, dev_name)
        return

    def __append_usb_ups_info(self, node, dev_name):
        """Append usb UPS info to specified node"""

        if self.__is_non_vendor_ups_device(dev_name):
            ups_name = self.__get_real_ups_device(dev_name)
        else:
            ups_name = dev_name

        self.__append_usb_info(node, ups_name)
        return

    def __append_query_info(self, dev_node):
        """Append query info to specified node"""

        dev_name = dev_node.get_node(NAME_NAME).get_text()
        dev_type = self.__get_device_type(dev_name)

        append_info_func = {
            SATA_HDD_TYPE: self.__append_sata_hdd_info,
            USB_HDD_TYPE: self.__append_usb_hdd_info,
            USB_LPR_TYPE: self.__append_usb_lpr_info,
            USB_UPS_TYPE: self.__append_usb_ups_info,
        }

        append_info_func[dev_type](dev_node, dev_name)
        return

    def __is_non_vendor_ups_device(self, dev_name):

        if dev_name in UPS_DEV_LIST:
            return True
        return False

    def __get_real_ups_device(self, ups_name):

        for real_ups_name in REAL_USB_UPS_LIST + REAL_SERIAL_UPS_LIST:

            if not real_ups_name.startswith(ups_name):
                continue

            if os.path.exists(os.path.join(DEV_DIR, real_ups_name)):
                return real_ups_name

        return ""

    def __is_ups_device_exist(self, ups_name):

        for real_ups_name in REAL_USB_UPS_LIST + REAL_SERIAL_UPS_LIST:

            if not real_ups_name.startswith(ups_name):
                continue

            if os.path.exists(os.path.join(DEV_DIR, real_ups_name)):
                return True

        return False

    #
    # Request handlers
    #

    def __query_handler(self, entry):
        """Handle "query" request. Get device info from tmp files."""

        self.__validate_root_tag_name(entry)

        try:
            self.__append_query_info(entry.get().get_node(DEVICE_NAME))
        except:
            entry.set_result(Results.FILE_NOT_FOUND)
            return entry

        entry.set_result(Results.OK)
        return entry

    def __chk_exist_handler(self, entry):
        """Handle "chk_exist" request.
           This handler check whether specified device exist.
        """

        self.__validate_root_tag_name(entry)
        dev_name = entry.get().get_node(NAME_PATH).get_text()

        if self.__is_non_vendor_ups_device(dev_name):
            if self.__is_ups_device_exist(dev_name):
                entry.set_result(Results.OK)
            else:
                entry.set_result(Results.DEVICE_NOT_FOUND)

        else:
            dev_path = util.symbl_dev_name2dev_path(dev_name)

            if os.path.exists(dev_path):
                entry.set_result(Results.OK)
            else:
                entry.set_result(Results.DEVICE_NOT_FOUND)

        return entry

    def __settle_handler(self, entry):
        """Handle "settle" request."""

        CmdExec(list(UDEV_SETTLE_ARG),
                stdout=NULL, stderr=NULL).execute()
        entry.set_result(Results.OK)
        return entry

    def __list_handler(self, entry):
        """Handle "list" request."""

        ROOT_XML = "<devctrl><devicelist/></devctrl>"
        DEVICE_NODE_XML = "<device><name>%s</name></device>"

        root_node = XmlNode(ROOT_XML)
        devlist_node = root_node.get_node(DEVICELIST_NAME)

        for dev_name in ALL_REAL_DEV_LIST:

            dev_node = XmlNode(DEVICE_NODE_XML % dev_name)

            if not os.path.islink(os.path.join(DEV_DIR, dev_name)):
                continue

            try:
                self.__append_query_info(dev_node)
            except:
                continue

            devlist_node.append_child(dev_node)

        entry.set(root_node)
        entry.set_result(Results.OK)
        return entry

    #
    # Constractor
    #

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.CHK_EXIST, self.__chk_exist_handler)
        self.set_handler(ReqTypes.SETTLE, self.__settle_handler)
        self.set_handler(ReqTypes.QUERY, self.__query_handler)
        self.set_handler(ReqTypes.LIST, self.__list_handler)

#
# Main
#

if __name__ == "__main__":
    PalDevice(option=palcmd.PalCmdOption()).handle_request()
