#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for DiskCtrl."""

from __future__ import with_statement

import os
import time
import re
import pal_config as config
import libpal.palcmd as palcmd
from pal_info import ReqTypes
from pal_info import Results
from iobase.cmdexec import (CmdExec, PIPE, NULL)
from iobase.exception import (InvalidArgError, NotFoundError, ExecError)
from iobase.xmldata import (XmlNode)
from pal_config import (SMARTCTL_CMD, PARTED_CMD, DD_CMD, BLOCKDEV_CMD,
                        UDEVADM_CMD, MOUNT_CMD, UMOUNT_CMD, SG_START_CMD)
from libpal.util import get_sata_disk_info
DISKLABEL_CMD = "dummy"  # fixme

from pal_config import (DISK_X_MODEL_FILE, DISK_X_REV_FILE,
                        DISK_X_VENDOR_FILE, DISK_X_SIZE_FILE)
from pal_config import (DEV_DIR)
import ionas.device as device

### XML Tag Name ###
DISKCTRL_NAME = "diskctrl"
DISK_NAME = "disk"
NAME_NAME = "name"
MODEL_NAME = "model"
REV_NAME = "rev"
VENDOR_NAME = "vendor"
SIZE_NAME = "size"
NUMBER_NAME = "number"
PARTTYPE_NAME = "parttype"
FILESYSTEM_NAME = "filesystem"
START_NAME = "start"
END_NAME = "end"
SMART_STATUS_NAME = "smartstatus"
DISK_ERR_CNT_NAME = "diskerrcnt"
PARTITIONLIST_NAME = "partitionlist"
PARTITION_NAME = "partition"
OPTION_NAME = "option"
DISK_ERR_CNT_NAME = "diskerrcnt"
OPERATION_NAME = "operation"
SPIN_UP_NAME = "spinup"
SPIN_DOWN_NAME = "spindown"

### XML Attribute ###
CLEAR_ATTR = "clear"

### XML Attribute Value ###
PTABLE_ATTR = "ptable"

### XML Element ###
SMART_ERROR_EL = "error"
SMART_NORMAL_EL = "noerror"
NOERROR_EL = "noerror"
ERROR_EL = "error"
MAX_END_LBA_EL = "-1"

DISK_NAME_XPATH = "./disk/name"

### OPTION ###
OPERATION_OPT = "operation"

### Command and argment ###
#
# XXXXX_ARG_FORMAT is the prototype list of the command which inclues VAR.
# VAR is replaced dynamically in __execute_cmd() method using
# XXXXX_VAR_INDEX_LIST and some variables._FORMAT
#

VAR = "%s"
# FORMAT: disklabel --check [disk path]
DISKLABEL_CHK_ARG_FORMAT = [DISKLABEL_CMD, "--check", VAR]
DISKLABEL_CHK_VAR_INDEX_LIST = [2]

# FORMAT: disklabel [disk path] errorcount
DISKLABEL_ERRCNT_ARG_FORMAT = [DISKLABEL_CMD, VAR, "errorcount"]
DISKLABEL_ERRCNT_VAR_INDEX_LIST = [1]

# FORMAT: smartctl --smart=on [disk path]
ENABLE_SMART_ARG_FORMAT = [SMARTCTL_CMD, "--smart=on", VAR]
ENABLE_SMART_VAR_INDEX_LIST = [2]

# FORMAT: smartctl --saveauto=on [disk path]
ENABLE_SAVEAUTO_ARG_FORMAT = [SMARTCTL_CMD, "--saveauto=on", VAR]
ENABLE_SAVEAUTO_VAR_INDEX_LIST = [2]

# FORMAT: smartctl --health [disk path]
SMART_ERROR_CHECK_ARG_FORMAT = [SMARTCTL_CMD, "--health", VAR]
SMART_ERROR_CHECK_VAR_INDEX_LIST = [2]

# FORMAT: parted -s [disk path] mklabel [partition type]
MKLABEL_ARG_FORMAT = [PARTED_CMD, "-s", VAR, "mklabel", VAR]
MKLABEL_VAR_INDEX_LIST = [2, 4]

# FORMAT: parted -s [disk path] unit s mkpart
#             primary [file system] [startpos] [endpos]
MKPART_ARG_FORMAT = [PARTED_CMD, "-s", VAR, "unit", "s", "mkpart",
                     "primary", VAR, VAR, VAR]
MKPART_VAR_INDEX_LIST = [2, 7, 8, 9]

# FORMAT: parted -s [disk path] unit s mkpart
#             [name] [file system] [startpos] [endpos]
MKPART_GPT_ARG_FORMAT = [PARTED_CMD, "-s", VAR, "unit", "s", "mkpart",
                         VAR, VAR, VAR, VAR]
MKPART_GPT_VAR_INDEX_LIST = [2, 6, 7, 8, 9]

# FORMAT: parted -s [disk path] rm [partition number]
DELPART_ARG_FORMAT = [PARTED_CMD, "-s", VAR, "rm", VAR]
DELPART_VAR_INDEX_LIST = [2, 4]

# FORMAT: dd if=/dev/zero of=[partition path] bs=4096 count=16
ZERO_CLR_PART_ARG_FORMAT = \
    [DD_CMD, "if=/dev/zero", "of=%s", "bs=4096", "count=16"]
ZERO_CLR_PART_VAR_INDEX_LIST = [2]

# FORMAT: parted VAR unit s p
GET_PARTITION_ARG_FORMAT = [PARTED_CMD, VAR, "unit", "s", "print"]
GET_PARTITION_VAR_INDEX_LIST = [1]

# FORMAT: blockdev [disk path]
REREADPT_ARG_FORMAT = [BLOCKDEV_CMD, "--rereadpt", VAR]
REREADPT_VAR_INDEX_LIST = [2]

# FORMAT: smartctrl -A [disk path]
GET_SMARTDATA_ARG_FORMAT = [SMARTCTL_CMD, "-A", VAR]
GET_SMARTDATA_INDEX_LIST = [2]

# FORMAT: smartctrl -i [disk path]
GET_IDENTIFY_ARG_FORMAT = [SMARTCTL_CMD, "-i", VAR]
GET_IDENTIFY_INDEX_LIST = [2]

UDEV_SETTLE_ARG = [UDEVADM_CMD, "settle"]

# FORMAT: dd if=[disk path] of=/dev/zero bs=512 skip=%s count=%s
ZEROCLEAR_ARG_FORMAT = [DD_CMD, "if=/dev/zero", "of=%s",
                        "bs=512", "skip=%s", "count=%s"]
ZEROCLEAR_INDEX_LIST = [2, 4, 5]

# FORMAT: mount -t [type] [partition path] [destination pth]
MOUNT_NON_OPTION_ARG_FORMAT = [MOUNT_CMD, "-t", VAR, VAR, VAR]
MOUNT_NON_OPTION_VAR_INDEX_LIST = [2, 3, 4]

# FORMAT: umount [destination pth]
UNMOUNT_ARG_FORMAT = [UMOUNT_CMD, VAR]
UNMOUNT_VAR_INDEX_LIST = [1]

# FORMAT: sg_start --start [device]
SPIN_UP_ARG_FORMAT = [SG_START_CMD, "--start", VAR]
SPIN_UP_VAR_INDEX_LIST = [2]

# FORMAT: sg_start --stop [device]
SPIN_DOWN_ARG_FORMAT = [SG_START_CMD, "--stop", "--readonly", VAR]
SPIN_DOWN_VAR_INDEX_LIST = [3]


### Some value ###
GET_SMART_ERR_MASK = 0x08

GPT_TABLE_SIZE = 64

MODEL_PREFIX_VENDOR_MAP = \
    {"ST": "Seagate",
     "WD": "Western Digital",
     "Hitachi": "Hitachi",
     "SAMSUNG": "Samsung"}

RANGE_TO_TAG_MAP = {
    "0-20ms": "range0_20ms",
    "20-40ms": "range20_40ms",
    "40-60ms": "range40_60ms",
    "60-80ms": "range60_80ms",
    "80-100ms": "range80_100ms",
    "100-120ms": "range100_120ms",
    "120-140ms": "range120_140ms",
    "140-160ms": "range140_160ms",
    "160-180ms": "range160_180ms",
    "180-200ms": "range180_200ms",
    "200ms-500ms": "range200_500ms",
    "500ms-1s": "range500_1s",
    "1s-5s": "range1_5s",
    "5s-10s": "range5_10s",
    "10s-15s": "range10_15s",
    "15s-": "range15s_",
}

STATISTICS_BASE_XML = \
    """<statistics>
          <signature/>
          <read><total/></read>
          <write><total/></write>
        </statistics>"""


class PalDiskCtrl(palcmd.PalCmd):
    """Pal DiskCtrlcommand implementation class."""

    class __MakePartitionError(Exception):
        """Make Partition Error."""

        pass

    class __DeletePartitionError(Exception):
        """Delete Partition Error."""

        pass

    class __MakeLabelError(Exception):
        """Make Disk Label Error."""

        pass

    class __ClearError(Exception):
        """Clear Error."""

    class __MountError(Exception):
        "Mount Error"
        pass

    class __UnmountError(Exception):
        "Unmount Error"
        pass

    class __DetectError(Exception):
        "Detect Error"
        pass

    class __DeviceNotFound(Exception):
        "Device Not Found"
        pass

    class __NotSupported(Exception):
        "Not Supported"
        pass

    #
    # Sub functions
    #

    def __validate_root_tag_name(self, entry):
        """Validate the root tag name."""

        root_tag_name = entry.get().get_node_name()
        if DISKCTRL_NAME != root_tag_name:
            raise InvalidArgError("invalid root tag..")

    def __is_exist_node(self, node, node_path):
        """Check whether node_path exists in specified node"""

        try:
            node.get_node(node_path)
        except NotFoundError:
            return False
        return True

    def __load_value(self, file_name):
        """load value from specified file"""

        with open(file_name) as fh:
            return fh.read().rstrip()

    def __force_make_dirs(self, dir_path):
        """make specified directory."""

        if not os.path.isdir(dir_path):
            os.makedirs(dir_path)
        return

    def __convert_real_devname(self, dev_name):
        """Convert synbol device name to real device name.
           If dev_name is already real this method just returns it.
        """

        dev_path = os.path.join(DEV_DIR, dev_name)
        return os.path.basename(os.path.realpath(dev_path))

    def __convert_real_devpath(self, dev_name):
        """Convert synbol device name to real device name.
           If dev_name is already real this method just returns it.
        """

        return os.path.realpath(os.path.join(DEV_DIR, dev_name))

    def __execute_cmd(self, is_enable_exception, is_wait,
                      std_out, std_err,
                      cmd_fmt, cmd_var_idx_list=[], variable_list=[]):
        """Execute command and return Popen object
              cmd_fmt - specify command format list
              cmd_var_idx_list - specify variable index of cmd_fmt
              variable_list - variables corresponding to cmd_var_idx_list
        """

        cmd = cmd_fmt[:]

        if len(cmd_var_idx_list) != len(variable_list):
            raise ExecError("Command list size is bad.. (__execute_cmd)")

        for (loop_idx, var_idx) in enumerate(cmd_var_idx_list):
            cmd[var_idx] = cmd[var_idx] % variable_list[loop_idx]

        return CmdExec(cmd,
                       stdout=std_out,
                       stderr=std_err,
                       wait=is_wait,
                       raise_exc=is_enable_exception).execute()

    def __is_smart_error(self, disk_path):
        """Check SMART error"""

        if os.path.basename(disk_path).startswith("usbd"):
            return False

        self.__execute_cmd(True, True, NULL, NULL,
                           ENABLE_SMART_ARG_FORMAT,
                           ENABLE_SMART_VAR_INDEX_LIST,
                           [disk_path])

        self.__execute_cmd(True, True, NULL, NULL,
                           ENABLE_SAVEAUTO_ARG_FORMAT,
                           ENABLE_SAVEAUTO_VAR_INDEX_LIST,
                           [disk_path])

        proc_smart_chk = self.__execute_cmd(False, True, NULL, NULL,
                                            SMART_ERROR_CHECK_ARG_FORMAT,
                                            SMART_ERROR_CHECK_VAR_INDEX_LIST,
                                            [disk_path])

        if proc_smart_chk.returncode & GET_SMART_ERR_MASK:
            return True
        else:
            return False

    def __get_partition_list_node(self, disk_path):
        """Get partitionlist node."""

        proc = self.__execute_cmd(False, False, PIPE, NULL,
                                  GET_PARTITION_ARG_FORMAT,
                                  GET_PARTITION_VAR_INDEX_LIST,
                                  [disk_path])

        part_info_list = []
        part_table = None
        name_idx = 0
        for line in  proc.stdout.readlines():

            line = line.rstrip()

            if part_table is None:
                match = re.match("Partition Table:\s*(\S+)", line)
                if match:
                    part_table = match.group(1)
                continue

            if part_table == "gpt":

                match = re.match("Number\s+"
                                 "Start\s+"
                                 "End\s+"
                                 "Size\s+"
                                 "File system\s+"
                                 "Name\s+"
                                 "Flags",
                                 line)
                if match:
                    name_idx = line.find("Name")
                    continue

                match = re.match("\s+([0-9]+)"
                                 "\s+([0-9]+)s"
                                 "\s+([0-9]+)s"
                                 "\s+[0-9]+s",
                                 line)
                if not match:
                    continue

                # Partition Geometory
                part_info_dict = {NUMBER_NAME: match.group(1),
                                  START_NAME: match.group(2),
                                  END_NAME: match.group(3)}

                # Partition Name
                match = re.match("((\S+\s?)+\S)", line[name_idx:])
                if match:
                    part_info_dict[NAME_NAME] = match.group(1)

                part_info_list.append(part_info_dict)
                continue

            if part_table == "msdos":

                match = re.match("\s+([0-9]+)"
                                 "\s+([0-9]+)s"
                                 "\s+([0-9]+)s"
                                 "\s+[0-9]+s"
                                 "\s+(\S*)",
                                 line)
                if not match:
                    continue

                part_info_list.append({NUMBER_NAME: match.group(1),
                                       START_NAME: match.group(2),
                                       END_NAME: match.group(3),
                                       PARTTYPE_NAME: match.group(4)})
                continue

        partlist_node = XmlNode(PARTITIONLIST_NAME)

        if part_table is not None:
            partlist_node.set_attrs({PTABLE_ATTR: part_table})

        for part_info in part_info_list:
            part_node = XmlNode(PARTITION_NAME)

            for tag_name, value in part_info.items():
                part_node.append_child(tag_name, value)

            partlist_node.append_child(part_node)

        return partlist_node

    def __do_udev_settle(self):
        """Excecute udev settle"""

        try:
            self.__execute_cmd(False, True, NULL, NULL,
                               UDEV_SETTLE_ARG)
        except:
            pass

        return

    def __reread_patrition_table(self, disk_path):
        """Re-read partition table.
           After mkfs this method is needed to call.
        """

        self.__execute_cmd(True, False, PIPE, NULL,
                           REREADPT_ARG_FORMAT,
                           REREADPT_VAR_INDEX_LIST,
                           [disk_path])
        time.sleep(2)
        self.__do_udev_settle()

        try:
            self.dev_mgr.save_dev_info(os.path.basename(disk_path))
        except:
            self.exception("fail to save_dev_info")
            pass

        return

    def __clear_disk(self, disk_path):
        """Clear disk."""

        self.__execute_cmd(True, True, NULL, NULL,
                           ZERO_CLR_PART_ARG_FORMAT,
                           ZERO_CLR_PART_VAR_INDEX_LIST,
                           [disk_path])

        return True

    def __delete_partition(self, disk_node, is_clear):
        """Delete partition by disk node info"""

        NUMBER_XPATH = "./partitionlist/partition/number"
        disk_path = os.path.join(DEV_DIR,
                                 disk_node.get_node(NAME_NAME).get_text())

        if is_clear:
            self.__clear_disk(disk_path)
            self.__reread_patrition_table(disk_path)
            return

        num_node_list = disk_node.get_node_list(NUMBER_XPATH)
        for num_node in num_node_list:
            part_num = num_node.get_text()

            proc = self.__execute_cmd(False, True, NULL, NULL,
                                      DELPART_ARG_FORMAT,
                                      DELPART_VAR_INDEX_LIST,
                                      [disk_path, part_num])
            if 0 != proc.returncode:
                raise self.__DeletePartitionError
        return

    def __make_gpt_partition(self, disk_path, part_node):
        """Make GPT partition"""

        part_label = part_node.get_node(NAME_NAME).get_text()
        start_pos = part_node.get_node(START_NAME).get_text()
        end_pos = part_node.get_node(END_NAME).get_text()

        proc = self.__execute_cmd(False, True, NULL, NULL,
                                  MKPART_GPT_ARG_FORMAT,
                                  MKPART_GPT_VAR_INDEX_LIST,
                                  [disk_path, part_label, "xfs",
                                   start_pos, end_pos])
        if 0 != proc.returncode:
            raise self.__MakePartitionError

        return

    def __make_mbr_partition(self, disk_path, part_node):
        """Make MBR partition"""

        filesystem = part_node.get_node(FILESYSTEM_NAME).get_text()

        if "fat" in filesystem:
            filesystem = "fat32"
        else:
            filesystem = "ext3"

        start_pos = part_node.get_node(START_NAME).get_text()
        end_pos = part_node.get_node(END_NAME).get_text()

        proc = self.__execute_cmd(False, True, NULL, NULL,
                                  MKPART_ARG_FORMAT,
                                  MKPART_VAR_INDEX_LIST,
                                  [disk_path, filesystem, start_pos, end_pos])
        if 0 != proc.returncode:
            raise self.__MakePartitionError

        return

    def __make_partition(self, disk_node):
        """Make partition"""

        PARTITION_XPATH = "./partitionlist/partition"

        disk_path = os.path.join(DEV_DIR,
                                 disk_node.get_node(NAME_NAME).get_text())
        real_disk_path = os.path.realpath(disk_path)
        part_table = \
            disk_node.get_node(PARTITIONLIST_NAME).get_attr_value(PTABLE_ATTR)

        part_node_list = disk_node.get_node_list(PARTITION_XPATH)

        for part_node in part_node_list:

            if MAX_END_LBA_EL == part_node.get_value(END_NAME):

                disk_size = \
                    self.__load_value(DISK_X_SIZE_FILE % \
                        os.path.basename(real_disk_path))

                if part_table == "gpt":
                    end_sector = str(int(disk_size) - 1 - GPT_TABLE_SIZE)

                elif part_table == "msdos":
                    end_sector = str(int(disk_size) - 1)
                else:
                    raise InvalidArgError(
                              "invalid root part_table..(%s)" % part_table)

                part_node.get_node(END_NAME).set_text(end_sector)

            if part_table == "gpt":
                self.__make_gpt_partition(real_disk_path, part_node)
                continue

            if part_table == "msdos":
                self.__make_mbr_partition(real_disk_path, part_node)
                continue

        self.__reread_patrition_table(disk_path)
        return

    def __make_disklabel(self, disk_node):
        """Make disk label"""

        disk_path = os.path.join(DEV_DIR,
                                 disk_node.get_node(NAME_NAME).get_text())
        part_table = \
            disk_node.get_node(PARTITIONLIST_NAME).get_attr_value(PTABLE_ATTR)

        proc = self.__execute_cmd(False, True, NULL, NULL,
                                  MKLABEL_ARG_FORMAT,
                                  MKLABEL_VAR_INDEX_LIST,
                                  [disk_path, part_table])
        if 0 != proc.returncode:
            raise self.__MakeLabelError

        return

    def __add_partition(self, disk_node):

        PARTITION_XPATH = "./partitionlist/partition"

        disk_path = os.path.join(DEV_DIR,
                                 disk_node.get_node(NAME_NAME).get_text())
        real_disk_path = os.path.realpath(disk_path)

        part_node_list = disk_node.get_node_list(PARTITION_XPATH)

        for part_node in part_node_list:

            if MAX_END_LBA_EL == part_node.get_value(END_NAME):

                disk_size = \
                    self.__load_value(DISK_X_SIZE_FILE % \
                        os.path.basename(real_disk_path))
                disk_size = str(int(disk_size) - 100)
                part_node.get_node(END_NAME).set_text(disk_size)

            self.__make_gpt_partition(real_disk_path, part_node)

        self.__reread_patrition_table(disk_path)
        return

    def __append_basic_query_info(self, disk_node, is_get_partition_info=False):
        """Append query info to specified node"""

        disk_name = disk_node.get_node(NAME_NAME).get_text()
        real_disk_name = self.__convert_real_devname(disk_name)

        infofile_tag_pair = {MODEL_NAME: DISK_X_MODEL_FILE,
                             REV_NAME: DISK_X_REV_FILE,
                             VENDOR_NAME: DISK_X_VENDOR_FILE,
                             SIZE_NAME: DISK_X_SIZE_FILE}

        for tag_name, file_path in infofile_tag_pair.items():
            disk_node.append_child(tag_name,
                                   self.__load_value(file_path %
                                                     real_disk_name))

        disk_path = os.path.join(DEV_DIR, disk_name)

        if is_get_partition_info:
            disk_node.append_child(self.__get_partition_list_node(disk_path))
        return

    def __get_atainfo_node(self, disk_path):

        ATA_INFO_NAME = "atainfo"
        atainfo_node = XmlNode(ATA_INFO_NAME)
        info_dict = get_sata_disk_info(disk_path)
        for key, value in info_dict.iteritems():
            atainfo_node.append_child(key, value)

        return atainfo_node

    def __append_ata_info(self, disk_node):

        disk_name = disk_node.get_node(NAME_NAME).get_text()

        atainfo_node = \
            self.__get_atainfo_node(self.__convert_real_devpath(disk_name))

        disk_node.append_child(atainfo_node)
        return

    def __get_smartctl_output(self, disk_path):
        """Return 'smartctl' output."""

        proc = self.__execute_cmd(False, False, PIPE, NULL,
                                  GET_SMARTDATA_ARG_FORMAT,
                                  GET_SMARTDATA_INDEX_LIST,
                                  [disk_path])
        return proc.stdout

    def __get_smartdatalist_node(self, disk_path):
        "Return the 'smartdatalist' tree transformed from 'smartctl' output."

        SUBELEMENT_LIST = (
                            ('id', 1),
                            ('flag', 2),
                            ('value', 3),
                            ('worstvalue', 4),
                            ('threshold', 5),
                            ('rawvalue', 6),
                          )

        smartdata_re = re.compile('\s*([0-9]+)\s*\S+\s+(0x[0-9a-fA-F]+)\s+'
                            '([0-9]+)\s+([0-9]+)\s+([0-9]+)\s+\S+\s+\S+\s+'
                            '\S+\s+([0-9]+)')

        smartdatalist_node = XmlNode("smartdatalist")
        for line in self.__get_smartctl_output(disk_path):
            match = smartdata_re.match(line)
            if not match:
                continue

            smartdata_node = XmlNode("smartdata")
            for (name, value) in SUBELEMENT_LIST:
                smartdata_node.append_child(name, match.group(value).strip())
            smartdatalist_node.append_child(smartdata_node)

        return smartdatalist_node

    def __append_smartinfo(self, disk_node):

        disk_name = disk_node.get_node(NAME_NAME).get_text()

        smartinfo_node = XmlNode("smartinfo")
        smartdatalist_node = \
            self.__get_smartdatalist_node(\
                self.__convert_real_devpath(disk_name))
        smartinfo_node.append_child(smartdatalist_node)

        disk_node.append_child(smartinfo_node)
        return

    def __get_statistics_node(self, disk_name):

        def get_re_match_str(re_str, target_str):

            match = re.match(re_str, target_str)
            if match:
                return match.group(1)
            return None

        PROC_RESPONSE_DIR = '/proc/scsiinfo/response'

        SIGNATURE_RE = "signature:\s*([0-9a-zA-Z]+)"
        READ_TOTAL_RE = "Read commands total:\s*([0-9]+)"
        WRITE_TOTAL_RE = "Write commands total:\s*([0-9]+)"
        RANGE_RE = "(.*)\s*:\s*([0-9]+)"

        resp_node = XmlNode(STATISTICS_BASE_XML)

        kdev_name = self.dev_mgr.get_kdev_name(disk_name)

        info_file = \
            open(os.path.join(PROC_RESPONSE_DIR, kdev_name))

        signature = None
        read_cmd_total = None
        write_cmd_total = None
        is_read_line = False
        is_write_line = False

        for line in info_file:

            if not signature:
                signature = get_re_match_str(SIGNATURE_RE, line)
                if signature:
                    resp_node.get_node("signature").set_text(signature)
                    continue

            if not read_cmd_total:
                read_cmd_total = get_re_match_str(READ_TOTAL_RE, line)
                if read_cmd_total:
                    resp_node.get_node("read/total").set_text(read_cmd_total)
                    is_read_line, is_write_line = True, False
                    continue

            if not write_cmd_total:
                write_cmd_total = get_re_match_str(WRITE_TOTAL_RE, line)
                if write_cmd_total:
                    resp_node.append_child("write")
                    resp_node.get_node("write/total").set_text(write_cmd_total)
                    is_read_line, is_write_line = False, True
                    continue

            # response time range
            match = re.match(RANGE_RE, line)
            if not match:
                continue
            range_name = match.group(1)
            range_cnt = match.group(2)
            tag_name = RANGE_TO_TAG_MAP.get(range_name)

            if not tag_name:
                continue

            if is_read_line:
                resp_node.get_node("read").append_child(tag_name, range_cnt)
                continue

            if is_write_line:
                resp_node.get_node("write").append_child(tag_name, range_cnt)
                continue

        return resp_node

    def __append_disk_statistics_info(self, disk_node):

        disk_name = disk_node.get_node(NAME_NAME).get_text()
        disk_node.append_child(self.__get_statistics_node(disk_name))
        return

    def __zero_clear_disk(self, disk_name, start_sector, end_sector):

        skip = str(start_sector)
        count = str(end_sector - start_sector)
        disk_path = os.path.realpath(os.path.join(DEV_DIR, disk_name))

        proc = self.__execute_cmd(False, True, NULL, NULL,
                                  ZEROCLEAR_ARG_FORMAT,
                                  ZEROCLEAR_INDEX_LIST,
                                  [disk_path, skip, count])
        if 0 != proc.returncode:
            raise self.__ClearError

        try:
            self.__reread_patrition_table(os.path.join(DEV_DIR, disk_name))
        except:
            self.exception("fail to __reread_patrition_table")
            pass

        return

    def __mount_partition(self, fstype, src, dst):

        if not os.path.isdir(dst):
            os.makedirs(dst)

        proc = self.__execute_cmd(False, True,
                                  NULL, NULL,
                                  MOUNT_NON_OPTION_ARG_FORMAT,
                                  MOUNT_NON_OPTION_VAR_INDEX_LIST,
                                  [fstype, src, dst])
        if 0 != proc.returncode:
            raise self.__MountError

    def __unmount_partition(self, mnt_point):

        proc = self.__execute_cmd(False, True,
                                  NULL, NULL,
                                  UNMOUNT_ARG_FORMAT,
                                  UNMOUNT_VAR_INDEX_LIST,
                                  [mnt_point])
        if 0 != proc.returncode:
            raise self.__UnmountError

    def __get_error_count(self, disk_name):

        DISK_LABEL_PART = "p4"
        DISK_LABEL_FSTYPE = "ext4"
        DISK_LABEL_DIR = "disklabel"
        TMP_DISK_LABEL_DIR = os.path.join(config.TEMP_DIR,
                                          "tmpdisklabel")
        ERRCNT_FILE = "errorcount"

        if not os.path.isdir(TMP_DISK_LABEL_DIR):
            os.makedirs(TMP_DISK_LABEL_DIR)

        part_path = os.path.join(DEV_DIR, disk_name + DISK_LABEL_PART)
        if not os.path.exists(part_path):
            raise self.__DeviceNotFound

        part_path = os.path.realpath(part_path)

        errcnt = 0
        try:
            self.__mount_partition(DISK_LABEL_FSTYPE,
                                   part_path,
                                   TMP_DISK_LABEL_DIR)

            errcnt_file_path = os.path.join(TMP_DISK_LABEL_DIR,
                                            DISK_LABEL_DIR,
                                            ERRCNT_FILE)

            if os.path.exists(errcnt_file_path):
                with open(errcnt_file_path) as fh:
                    errcnt = int(fh.read().rstrip())
        except:
            pass

        finally:
            try:
                self.__unmount_partition(TMP_DISK_LABEL_DIR)
            except:
                self.exception("fail to __simple_mount")

        return errcnt

    def __delete_device_from_system(self, dev_name):

        real_dev_name = self.dev_mgr.get_kdev_name(dev_name)

        delete_file_path = config.DEVICE_DELETE_FILE_TMPL % real_dev_name
        if not os.path.exists(delete_file_path):
            return

        with open(config.DEVICE_DELETE_FILE_TMPL % real_dev_name, "w") as fh:
            fh.write("1")

        self.__do_udev_settle()

        return

    def __delete_dev_node(self, dev_name):

        self.dev_mgr.delete_dev_info(dev_name,
                                     only_not_exists=True,
                                     for_all=True)
        return

    #
    # Request handlers
    #

    def __query_handler(self, entry):
        """Handle "query" request.
           This function get information from each temporary files.
        """

        DISK_NAME_XPATH = "./disk/name"

        ATA_ATTR = "ata"
        GETPART_ATTR = "getpartitioninfo"
        is_ata = False
        is_get_partition_info = False
        if self.__is_exist_node(entry.get(), OPTION_NAME):
            opt_node = entry.get().get_node(OPTION_NAME)
            is_ata = opt_node.is_set_to_true(ATA_ATTR)
            is_get_partition_info = opt_node.is_set_to_true(GETPART_ATTR)

        self.__validate_root_tag_name(entry)

        disk_path = \
            os.path.join(DEV_DIR, entry.get().get_value(DISK_NAME_XPATH))
        if not os.path.exists(disk_path):
            entry.set_result(Results.DEVICE_NOT_FOUND)
            return entry

        self.__append_basic_query_info(entry.get().get_node(DISK_NAME),
                                       is_get_partition_info)

        if is_ata:
            self.__append_ata_info(entry.get().get_node(DISK_NAME))
            self.__append_smartinfo(entry.get().get_node(DISK_NAME))
            self.__append_disk_statistics_info(entry.get().get_node(DISK_NAME))

        entry.set_result(Results.OK)
        return entry

    def __check_handler(self, entry):
        """Handle "check" request.
           This function check SMART and Disk error count.
        """

        ERROR_CNT_THRESHOLD = 100

        self.__validate_root_tag_name(entry)

        disk_node = entry.get().get_node(DISK_NAME)
        disk_name = disk_node.get_node(NAME_NAME).get_text()
        disk_path = os.path.join(DEV_DIR, disk_name)

        if not os.path.exists(disk_path):
            entry.set_result(Results.DEVICE_NOT_FOUND)
            return entry

        # smart
        if self.__is_smart_error(disk_path):
            disk_node.append_child(SMART_STATUS_NAME, SMART_ERROR_EL)
        else:
            disk_node.append_child(SMART_STATUS_NAME, SMART_NORMAL_EL)

        # errorcount
        errcnt = 0
        try:
            errcnt = self.__get_error_count(disk_name)
        except self.__DeviceNotFound:
            pass

        if errcnt >= ERROR_CNT_THRESHOLD:
            disk_node.append_child(DISK_ERR_CNT_NAME, ERROR_EL)
        else:
            disk_node.append_child(DISK_ERR_CNT_NAME, NOERROR_EL)

        entry.set_result(Results.OK)
        return entry

    def __mkpart_handler(self, entry):
        """Handle "mkpart" request."""

        self.__validate_root_tag_name(entry)

        disk_node = entry.get().get_node(DISK_NAME)

        try:
            self.__make_disklabel(disk_node)
            self.__make_partition(disk_node)

        except self.__MakePartitionError:
            entry.set_result(Results.CREATE_ERROR)

        except self.__MakeLabelError:
            entry.set_result(Results.CREATE_ERROR)

        else:
            entry.set_result(Results.OK)

        return entry

    def __addpart_handler(self, entry):

        self.__validate_root_tag_name(entry)
        disk_node = entry.get().get_node(DISK_NAME)

        try:
            self.__add_partition(disk_node)
        except self.__MakePartitionError:
            entry.set_result(Results.CREATE_ERROR)
        else:
            entry.set_result(Results.OK)

        return entry

    def __delpart_handler(self, entry):
        """Handle "mkpart" request."""

        self.__validate_root_tag_name(entry)

        is_clear = False

        if self.__is_exist_node(entry.get(), OPTION_NAME):
            opt_node = entry.get().get_node(OPTION_NAME)
            is_clear = opt_node.is_set_to_true(CLEAR_ATTR)

        try:
            self.__delete_partition(entry.get().get_node(DISK_NAME), is_clear)

        except self.__DeletePartitionError:
            entry.set_result(Results.DELETE_ERROR)

        else:
            entry.set_result(Results.OK)

        return entry

    def __clear_handler(self, entry):
        """Handle "clear" request."""

        self.__validate_root_tag_name(entry)

        disk_node = entry.get().get_node(DISK_NAME)

        disk_name = disk_node.get_node(NAME_NAME).get_text()
        start_sector = int(disk_node.get_node(START_NAME).get_text())
        end_sector = int(disk_node.get_node(END_NAME).get_text())

        try:
            self.__zero_clear_disk(disk_name,
                                   start_sector, end_sector)
        except self.__ClearError:
            entry.set_result(Results.CLEAR_ERROR)
        else:
            entry.set_result(Results.OK)

        return entry

    def __delete_handler(self, entry):
        """Handle "delete" request."""

        self.__validate_root_tag_name(entry)

        disk_name = entry.get().get_node(DISK_NAME_XPATH).get_text()

        is_fail = False

        try:
            self.__delete_device_from_system(disk_name)
        except:
            is_fail = True
            self.exception("fail to __delete_dev_node(%s) - do nothing" % \
                               disk_name)

        try:
            self.__delete_dev_node(disk_name)
        except:
            is_fail = True
            self.exception("fail to __delete_dev_node(%s) - do nothing" % \
                               disk_name)

        if is_fail:
            entry.set_result(Results.DELETE_ERROR)
        else:
            entry.set_result(Results.OK)

        return entry

    def __chk_exist_handler(self, entry):

        entry.set_result(Results.NOT_SUPPORTED)
        return entry

    def __spin_up(self, entry):

        disk_name = entry.get().get_node(DISK_NAME_XPATH).get_text()
        real_disk_path = self.__convert_real_devpath(disk_name)

        proc = self.__execute_cmd(False, True, NULL, NULL,
                                  SPIN_UP_ARG_FORMAT,
                                  SPIN_UP_VAR_INDEX_LIST,
                                  [real_disk_path])
        if 0 != proc.returncode:
            raise self.__StartError

    def __spin_down(self, entry):

        disk_name = entry.get().get_node(DISK_NAME_XPATH).get_text()
        real_disk_path = self.__convert_real_devpath(disk_name)

        proc = self.__execute_cmd(False, True, NULL, NULL,
                                  SPIN_DOWN_ARG_FORMAT,
                                  SPIN_DOWN_VAR_INDEX_LIST,
                                  [real_disk_path])
        if 0 != proc.returncode:
            raise self.__StopError

    def __control_handler(self, entry):

        self.__validate_root_tag_name(entry)

        ctrl_funcs = {SPIN_UP_NAME: self.__spin_up,
                      SPIN_DOWN_NAME: self.__spin_down}

        op = entry.get_options()[OPERATION_OPT]

        try:
            ctrl_funcs[op](entry)
            entry.set_result(Results.OK)
        except self.__StartError:
            entry.set_result(Results.START_ERROR)
        except self.__StopError:
            entry.set_result(Results.STOP_ERROR)

        return entry

    #
    # Constractor
    #

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.dev_mgr = device.DevMgr(config.DEV_DIR, config.DEV_LINK_DIR,
                                config.DEV_FILE_DIR, config.HDD_DEV_LIST)

        self.set_handler(ReqTypes.QUERY, self.__query_handler)
        self.set_handler(ReqTypes.CHECK, self.__check_handler)
        self.set_handler(ReqTypes.MKPART, self.__mkpart_handler)
        self.set_handler(ReqTypes.DELPART, self.__delpart_handler)
        self.set_handler(ReqTypes.CLEAR, self.__clear_handler)
        self.set_handler(ReqTypes.ADDPART, self.__addpart_handler)
        self.set_handler(ReqTypes.DELETE, self.__delete_handler)
        self.set_handler(ReqTypes.CHK_EXIST, self.__chk_exist_handler)
        self.set_handler(ReqTypes.CONTROL, self.__control_handler)

#
# Main
#

if __name__ == "__main__":
    PalDiskCtrl(option=palcmd.PalCmdOption()).handle_request()
