#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for LCM."""

import os
import re
import pal_config as config
import libpal.lcm.str_conv as conv
import libpal.palcmd as palcmd
import libpal.palcmd
from iobase.exception import (NotFoundError, InvalidArgError)
from iobase.cmdexec import (CmdExec, NULL)
from pal_config import (I2CSET_CMD)
from pal_info import ReqTypes
from pal_info import Results

### XML Tag Name ###
LCM_NAME = "lcm"
IMAGE_NAME = "image"
STRING_NAME = "string"
STRING_LIST_NAME = "stringlist"
ANY_CHILD_NAME = "*"

### XML Attribute ###
FILE_ATTR = "file"
LINE_ATTR = "line"

### Command and argment ###
#
# XXXXX_ARG_FORMAT is the prototype list of the command which inclues VAR.
# VAR is replaced dynamically in __execute_cmd() method using
# XXXXX_VAR_INDEX_LIST and some variables._FORMAT
#
VAR = "%s"

SET_IMAGE_AXIS_ARG = [I2CSET_CMD, "-y", "0", "0x8a", VAR, "w"]
SET_IMAGE_AXIS_VAR_INDEX_LIST = [4]

SET_IMAGE_BIT_PATTERN_ARG = [I2CSET_CMD, "-y", "0", "0x8d", VAR, "w"]
SET_IMAGE_BIT_PATTERN_VAR_INDEX_LIST = [4]

#i2cset -y 0 0xa [position] [character code] w
SET_CHARACTER_ARG = [I2CSET_CMD, "-y", "0", "0xa", VAR, VAR, "w"]
SET_CHARACTER_VAR_INDEX_LIST = [4, 5]

### LCM LINE horizontal position ###
LINE_1ST_POSITION = 0x10
LINE_2ND_POSITION = 0x20
LINE_MAX_CHARACTER = 0x09


class PalLcm(palcmd.PalCmd):
    """PalLcm command implementation class."""

    class __InvalidFile(Exception):
        """Invalid File Exception."""
        pass

    class __FileNotFound(Exception):
        """File Not Exception."""
        pass

    class __DisplayImpossible(Exception):
        """Display Impossible  Exception."""
        pass

    #
    # Sub functions
    #

    def __validate_root_tag_name(self, entry):
        """Validate the root tag name."""

        node_name = entry.get().get_node_name()
        if LCM_NAME != node_name:
            self.error("Invalid root node name %s" % node_name)
            raise InvalidArgError("invalid root tag..")

    def __get_value_by_key(self, target_key, dict):
        """This method is for utility.
           Find key of "dict" and return corresponding key.
        """

        for key, value in dict.items():
            if target_key == key:
                return value
        raise

    def __is_exist_attr(self, node, attr_val):
        """Check whether attr_val exists in specified node"""

        try:
            node.get_attr_value(attr_val)
        except NotFoundError:
            return False
        return True

    def __execute_cmd(self, is_enable_exception, is_wait,
                      std_out, std_err,
                      cmd_fmt, cmd_var_idx_list, *variables):
        """Execute command and return Popen object
              cmd_fmt - specify command format list
              cmd_var_idx_list - specify variable index of cmd_fmt
              variable_args - variables corresponding to cmd_var_idx_list
        """

        cmd = cmd_fmt[:]
        for (loop_idx, var_idx) in enumerate(cmd_var_idx_list):
            cmd[var_idx] = variables[loop_idx]

        proc = CmdExec(cmd, stdout=std_out, stderr=std_err,
                       raise_exc=is_enable_exception).execute()

        if is_wait:
            proc.wait()

        return proc

    def __display_stirng(self, lcm_string, line_num):
        """Display specified string to LCM."""

        lcm_code_list = conv.get_lcm_code_list(lcm_string)
        if len(lcm_code_list) > LINE_MAX_CHARACTER:
            raise self.__DisplayImpossible

        line_num_pos = {1: LINE_1ST_POSITION, 2: LINE_2ND_POSITION}
        if line_num not in line_num_pos:
            raise self.__DisplayImpossible
        lcm_position = line_num_pos[line_num]

        idx = -1
        for idx, lcm_code in enumerate(lcm_code_list):
            self.__execute_cmd(True, True, NULL, NULL,
                               SET_CHARACTER_ARG,
                               SET_CHARACTER_VAR_INDEX_LIST,
                               hex(lcm_position + idx),
                               hex(lcm_code))
        
        # 余った箇所にスペースを挿入
        for idx_space in range(idx+1, LINE_MAX_CHARACTER+1):
            self.__execute_cmd(True, True, NULL, NULL,
                               SET_CHARACTER_ARG,
                               SET_CHARACTER_VAR_INDEX_LIST,
                               hex(lcm_position + idx_space),
                               "0xa140") 
        return

    def __display_stirng_with_file(self, string_file_name, line_num=1):
        """Display string to LCM.
           Character data is read from specified file.
        """

        if not os.path.exists(string_file_name):
            raise self.__FileNotFound

        line_num_pos = {1: LINE_1ST_POSITION, 2: LINE_2ND_POSITION}
        if line_num not in line_num_pos:
            raise self.__DisplayImpossible
        lcm_position = line_num_pos[line_num]

        for idx, lcm_code in enumerate(open(string_file_name, "r")):

            if idx > LINE_MAX_CHARACTER:
                raise self.__InvalidFile

            lcm_code = lcm_code.rstrip()

            if "" == lcm_code:
                continue

            if re.match("^#.*", lcm_code):  # コメントアウト文字
                continue

            if not re.match("[0-9A-Fa-f]+$", lcm_code):
                raise self.__InvalidFile

            self.__execute_cmd(True, True, NULL, NULL,
                               SET_CHARACTER_ARG,
                               SET_CHARACTER_VAR_INDEX_LIST,
                               hex(lcm_position + idx),
                               hex(int(lcm_code, 16)))
        return

    def __transact_stringlist_node(self, lcm_node):
        """Transact 'stringlist' node"""

        string_list_node = lcm_node.get_node(STRING_LIST_NAME)
        string_node_list = string_list_node.get_node_list(STRING_NAME)

        for string_node in string_node_list:

            line_num = int(string_node.get_attr_value(LINE_ATTR))

            if self.__is_exist_attr(string_node, FILE_ATTR):

                str_file_path = string_node.get_attr_value(FILE_ATTR)
                self.__display_stirng_with_file(str_file_path, int(line_num))
                continue

            lcm_str = unicode(string_node.get_text(do_strip=False), "utf-8")
            self.__display_stirng(lcm_str, line_num)

        return

    #
    # Request handlers
    #

    def __display_handler(self, entry):
        """Handle "display" request."""

        self.__validate_root_tag_name(entry)

        try:
            self.__transact_stringlist_node(entry.get())

        except self.__InvalidFile:
            entry.set_result(Results.INVALID_FILE)
            return entry

        except self.__FileNotFound:
            entry.set_result(Results.FILE_NOT_FOUND)
            return entry

        except self.__DisplayImpossible:
            entry.set_result(Results.DISPLAY_ERROR)
            return entry

        entry.set_result(Results.OK)
        return entry

    #
    # Constractor
    #

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.DISPLAY, self.__display_handler)

#
# Main
#

if __name__ == "__main__":
    PalLcm(option=palcmd.PalCmdOption()).handle_request()
