#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for LED."""

from __future__ import with_statement

import libpal.palcmd as palcmd
from iobase.cmdexec import (CmdExec, NULL)
from iobase.exception import (ExecError, InvalidArgError, NotFoundError)
from pal_config import (I2CSET_CMD)
import pal_config as config
from pal_info import (Results, ReqTypes)

### XML Tag Name ###
LED_NAME = "led"
POWER_NAME = "power"
STATUS_NAME = "status"
HDD_NAME = "hdd"
BRIGHTNESS_NAME = "brightness"
ANY_CHILD_NAME = "*"

### XML Attribute ###
COLOR_ATTR = "color"
HDD_NUM_ATTR = "hddnum"

### XML Attribute value ###
HDD1_ATTR_VAL = "1"
HDD2_ATTR_VAL = "2"
HDD3_ATTR_VAL = "3"
HDD4_ATTR_VAL = "4"
HDD5_ATTR_VAL = "5"
HDD6_ATTR_VAL = "6"
RED = "red"
GREEN = "green"
BLUE = "blue"
ORANGE = "orange"
NON_COLER = "NOCOLOR"

### XML Element ###
ON_EL = "on"
BLINK_EL = "blink"
OFF_EL = "off"
BRIGHTNESS_LEVEL1_EL = "25"
BRIGHTNESS_LEVEL2_EL = "50"
BRIGHTNESS_LEVEL3_EL = "75"
BRIGHTNESS_LEVEL4_EL = "100"

STATUS_LED_GREEN = "led_green"
STATUS_LED_RED = "led_red"
STATUS_LED_ORANGE = "led_orange"

HDD1_LED_GREEN = "led_hdd1green"
HDD1_LED_RED = "led_hdd1red"

HDD2_LED_GREEN = "led_hdd2green"
HDD2_LED_RED = "led_hdd2red"

HDD3_LED_GREEN = "led_hdd3green"
HDD3_LED_RED = "led_hdd3red"

HDD4_LED_GREEN = "led_hdd4green"
HDD4_LED_RED = "led_hdd4red"

HDD5_LED_GREEN = "led_hdd4green"
HDD5_LED_RED = "led_hdd4red"

HDD6_LED_GREEN = "led_hdd4green"
HDD6_LED_RED = "led_hdd4red"

LED_ON = "on"
LED_BLINK = "blink"
LED_OFF = "off"

BRIGHTNESS_HIGH = "led_brightness high"
BRIGHTNESS_MIDDLE = "led_brightness middle"
BRIGHTNESS_LOW = "led_brightness low"
BRIGHTNESS_OFF = "led_brightness off"

WRITE_EEDATA_BRIGHTNESS_OFF_CMD = \
    [I2CSET_CMD, "-y", "0", "0x50", "0x20", "0x0"]

WRITE_EEDATA_BRIGHTNESS_LOW_CMD = \
    [I2CSET_CMD, "-y", "0", "0x50", "0x20", "0x3F"]

WRITE_EEDATA_BRIGHTNESS_MIDDLE_CMD = \
    [I2CSET_CMD, "-y", "0", "0x50", "0x20", "0x7F"]

WRITE_EEDATA_BRIGHTNESS_HIGH_CMD = \
    [I2CSET_CMD, "-y", "0", "0x50", "0x20", "0xFF"]


class PalLed(palcmd.PalCmd):
    """Pal Led command implementation class."""

    #
    # Sub functions
    #

    def __check_correct_root_node(self, entry):
        """Check whether root node is correctly.
        """

        node_name = entry.get().get_node_name()
        if node_name != LED_NAME:
            raise InvalidArgError(node_name)

    def __is_exist_node(self, node, node_path):
        """Check whether node_path exists in specified node"""

        try:
            node.get_node(node_path)
        except NotFoundError:
            return False
        return True

    def __get_value_by_key(self, target_key, dict):
        """This method is for utility.
           Find key of "dict" and return corresponding key.
        """

        for key, value in dict.iteritems():
            if target_key == key:
                return value
        raise InvalidArgError("Invalid key:" + str(target_key))

    def __execute_cmd(self, is_enable_exception, is_wait,
                      std_out, std_err,
                      cmd_fmt, cmd_var_idx_list=[], variable_list=[]):
        """Execute command and return Popen object
              cmd_fmt - specify command format list
              cmd_var_idx_list - specify variable index of cmd_fmt
              variable_list - variables corresponding to cmd_var_idx_list
        """

        cmd = cmd_fmt[:]

        if len(cmd_var_idx_list) != len(variable_list):
            raise ExecError("Command list size is bad.. (__execute_cmd)")

        for (loop_idx, var_idx) in enumerate(cmd_var_idx_list):
            cmd[var_idx] = cmd[var_idx] % variable_list[loop_idx]

        return CmdExec(cmd,
                       stdout=std_out,
                       stderr=std_err,
                       wait=is_wait,
                       raise_exc=is_enable_exception).execute()

    def __transact_status_led_node(self, status_led_node):
        """Transact status node.
        """

        try:
            color_attr_val = status_led_node.get_attr_value(COLOR_ATTR)
            tgt_color = self.__get_value_by_key(color_attr_val,
                                               {RED: STATUS_LED_RED,
                                                GREEN: STATUS_LED_GREEN,
                                                ORANGE: STATUS_LED_ORANGE})
        except NotFoundError:
            tgt_color = STATUS_LED_GREEN

        tgt_mode = \
            self.__get_value_by_key(status_led_node.get_text(),
                                    {ON_EL: LED_ON,
                                     BLINK_EL: LED_BLINK,
                                     OFF_EL: LED_OFF})

        for color in [STATUS_LED_RED, STATUS_LED_GREEN, STATUS_LED_ORANGE]:
            if tgt_color == color:
                mode = tgt_mode
            else:
                mode = LED_OFF

            with open(config.BTNLED, "w") as fh:
                fh.write(color + " " + mode)
        return

    def __get_hddnum_from_attr(self, target_key):
        """Get HDD number(integer) by specified key"""

        return self.__get_value_by_key(target_key,
                                       {HDD1_ATTR_VAL: 0,
                                        HDD2_ATTR_VAL: 1,
                                        HDD3_ATTR_VAL: 2,
                                        HDD4_ATTR_VAL: 3,
                                        HDD5_ATTR_VAL: 4,
                                        HDD6_ATTR_VAL: 5})

    def __get_cmd_hdd_map_cmd_opt(self, hdd_light_on_status_dict):

        map_opt_value = 0xf000
        for hdd_num, color_dict in hdd_light_on_status_dict.iteritems():

            if color_dict[RED] << hdd_num:
                map_opt_value += 0x01 << hdd_num * 2

            if color_dict[BLUE] << hdd_num:
                map_opt_value += 0x02 << hdd_num * 2
        return hex(int(map_opt_value))

    def __hold_hdd_led_color(self, hdd_num, color):

        REQ_COLOR_TO_HODL_COLOR = \
            {BLUE: config.HOLD_STATE_BLUE,
             RED: config.HOLD_STATE_RED,
             NON_COLER: config.HOLD_STATE_OFF}

        with open(config.HDD_LED_HOLD_STATE_FILE % hdd_num, "w") as fh:
            fh.write(REQ_COLOR_TO_HODL_COLOR[color])

    def __get_current_hdd_led_color(self, hdd_num):

        try:
            with open(config.HDD_LED_HOLD_STATE_FILE % hdd_num, "r") as fh:
                return fh.read().rstrip()
        except:
            return OFF_EL

    def __transact_hdd_led_status_node_list(self, hdd_led_status_node_list):
        """Transact ctrl hdd status led node.
           <hdd>
             <state color="red" hddnum="1" >on</state>
        """

        HDD_LED_STATE_MAP = \
            {0: {BLUE: HDD1_LED_GREEN, RED: HDD1_LED_RED},
             1: {BLUE: HDD2_LED_GREEN, RED: HDD2_LED_RED},
             2: {BLUE: HDD3_LED_GREEN, RED: HDD3_LED_RED},
             3: {BLUE: HDD4_LED_GREEN, RED: HDD4_LED_RED},
             4: {BLUE: HDD5_LED_GREEN, RED: HDD5_LED_RED},
             5: {BLUE: HDD6_LED_GREEN, RED: HDD6_LED_RED}}

        MODE_MAP = {ON_EL: LED_ON,
                    OFF_EL: LED_OFF,
                    BLINK_EL: LED_BLINK}

        HDD_LED_COLOR_LIST = [BLUE, RED]

        for status_node in hdd_led_status_node_list:

            # HDD number
            hddnum_attr_val = status_node.get_attr_value(HDD_NUM_ATTR)
            hdd_num = self.__get_hddnum_from_attr(hddnum_attr_val)

            # LED mode
            mode = status_node.get_text()

            # LED color
            try:
                tgt_color = status_node.get_attr_value(COLOR_ATTR)
                if OFF_EL == mode:
                    tgt_color = BLUE
            except NotFoundError:
                tgt_color = BLUE

            # LED制御
            with open(config.BTNLED, "w") as fh:
                fh.write(HDD_LED_STATE_MAP[hdd_num][tgt_color] + \
                         " " + MODE_MAP[mode])

            for color in HDD_LED_COLOR_LIST:
                if tgt_color == color:
                    continue

                with open(config.BTNLED, "w") as fh:
                    fh.write(HDD_LED_STATE_MAP[hdd_num][color] + \
                             " " + MODE_MAP[OFF_EL])

            # hold color to proc file
            #try:
            #    if OFF_EL == mode:
            #        self.__hold_hdd_led_color(hdd_num + 1, NON_COLER)
            #    else:
            #        self.__hold_hdd_led_color(hdd_num + 1, color)
            #except:
            #    self.exception("fail to __hold_hdd_led_color()")

        return

    def __get_hdd_led_brightness_cmdline(self, target_level):
        """Get the command line list that controls the
           brightness of HDD LED.
        """

        level_cmd_pair = \
            {BRIGHTNESS_LEVEL1_EL: BRIGHTNESS_OFF,
             BRIGHTNESS_LEVEL2_EL: BRIGHTNESS_LOW,
             BRIGHTNESS_LEVEL3_EL: BRIGHTNESS_MIDDLE,
             BRIGHTNESS_LEVEL4_EL: BRIGHTNESS_HIGH}

        return self.__get_value_by_key(target_level, level_cmd_pair)

    def __transact_hdd_brightness_node(self, hdd_led_brightness_node):
        """Transact HDD LED brightness node."""

        LEVEL_CMD_MAP = \
            {BRIGHTNESS_LEVEL1_EL: BRIGHTNESS_OFF,
             BRIGHTNESS_LEVEL2_EL: BRIGHTNESS_LOW,
             BRIGHTNESS_LEVEL3_EL: BRIGHTNESS_MIDDLE,
             BRIGHTNESS_LEVEL4_EL: BRIGHTNESS_HIGH}

        LEVEL_WRITE_EEDATA_CMD_MAP = \
            {BRIGHTNESS_LEVEL1_EL: WRITE_EEDATA_BRIGHTNESS_OFF_CMD,
             BRIGHTNESS_LEVEL2_EL: WRITE_EEDATA_BRIGHTNESS_LOW_CMD,
             BRIGHTNESS_LEVEL3_EL: WRITE_EEDATA_BRIGHTNESS_MIDDLE_CMD,
             BRIGHTNESS_LEVEL4_EL: WRITE_EEDATA_BRIGHTNESS_HIGH_CMD}

        brightness = hdd_led_brightness_node.get_text()
        cmd = LEVEL_CMD_MAP[brightness]
        with open(config.BTNLED, "w") as fh:
            fh.write(cmd)

        self.__execute_cmd(True, True, NULL, NULL,
                           LEVEL_WRITE_EEDATA_CMD_MAP[brightness])

    def __transact_hdd_led_node(self, hdd_led_node):
        """Transact hdd led node."""

        STATUS_XPATH = "statuslist/status"

        # HDD STATUS LED
        hdd_led_status_node_list = hdd_led_node.get_node_list(STATUS_XPATH)
        if len(hdd_led_status_node_list) > 0:
            self.__transact_hdd_led_status_node_list(hdd_led_status_node_list)

        return

    #
    # Request handlers
    #

    def __control_handler(self, entry):
        """Handle "control" request.
        """

        self.__check_correct_root_node(entry)

        target_led_node_list = entry.get().get_node_list(ANY_CHILD_NAME)
        for target_led_node in target_led_node_list:
            node_name = target_led_node.get_node_name()

            if STATUS_NAME == node_name:
                self.__transact_status_led_node(target_led_node)
                continue

            if HDD_NAME == node_name:
                self.__transact_hdd_led_node(target_led_node)
                continue

            if BRIGHTNESS_NAME == node_name:
                self.__transact_hdd_brightness_node(target_led_node)
                continue

        entry.set_result(Results.OK)
        return entry

    #
    # Constractor
    #

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.CONTROL, self.__control_handler)

#
# Main
#

if __name__ == "__main__":
    PalLed(option=palcmd.PalCmdOption()).handle_request()
