#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for Netatalk."""


import os
import re
import shutil
import errno
import glob

from libpal import (palcmd, util)
from pal_info import (ReqTypes, Results)
import iobase.xmldata as xmldata
from iobase.cmdexec import NULL
from iobase.exception import (NotFoundError, InvalidArgError)
from ionas.confut import (build_config_file, disable_setting, enable_setting)
import pal_config as config


USB_ORIG_SUFFIX = ".real"
# See "evh_ports"

SHARES_NAME = "shares"
NAME_NAME = "name"
TIMEMACHINE = "timemachine"
VOLUME_RE = r"/mnt/([^/]+)"

NAME_PATH = "share/name"
PATH_PATH = "share/path"
PERMISSION_PATH = "share/permission"
SHARESERVICE_PATH = "share/shareservice"
APPLESHARE_PATH = SHARESERVICE_PATH + "/appleshare"
READONLY_PATH = PERMISSION_PATH + "/readonly"
DETAILPERMISSION_PATH = PERMISSION_PATH + "/detailpermission"
READONLYMEMBER_PATH = DETAILPERMISSION_PATH + "/readonlymember"
RO_USER_LIST_PATH = READONLYMEMBER_PATH + "/userlist/*"
RO_GROUP_LIST_PATH = READONLYMEMBER_PATH + "/grouplist/*"
READWRITEMEMBER_PATH = DETAILPERMISSION_PATH + "/readwritemember"
RW_USER_LIST_PATH = READWRITEMEMBER_PATH + "/userlist/*"
RW_GROUP_LIST_PATH = READWRITEMEMBER_PATH + "/grouplist/*"
QUOTA_PATH = SHARESERVICE_PATH + "/msshare/quota"

OPTIONS_KEY = "options"
DBPATH_KEY = "dbpath"
ALLOW_KEY = "allow"
ROLIST_KEY = "rolist"
RWLIST_KEY = "rwlist"

TXTRECORD = "txt-record"
SERVICE = "service"
TXTRECORD_FMT_STR = "dk%s=adVN=%s,adVF=0x81"
TIMEMACHINE_SERVICE_STATE = '<?xml version="1.0" standalone=\'no\'?>' + \
                         '<!--*-nxml-*-->\n<!DOCTYPE service-group SYSTEM' + \
                         ' "avahi-service.dtd">\n\n'
TIMEMACHINE_SERVICE_FMT_STR = '<service-group><name replace-wildcards=' + \
                     '"yes">%%h</name><service><type>_adisk._tcp</type>' + \
                     '<port>9</port><txt-record>sys=adVF=0x100</txt-' + \
                     'record><txt-record>dk0=adVN=%s,adVF=0x81' + \
                     '</txt-record></service></service-group>'

WRITE_MODE = "w"
READ_MODE = "r"

GLOBAL_STR = "### default\n:DEFAULT: cnidscheme:dbd options:usedots," + \
        "upriv perm:0777 maccharset:MAC_JAPANESE volcharset:UTF8\n"
USER_FMT_STR = '%s'
GROUP_FMT_STR = '@%s'
AFP_CONF_COMMENT_FMT_STR = '# %s\n'
AFP_CONF_ITEM_FMT_STR_ONE = '"%s" "%s"'
AFP_CONF_ITEM_FMT_STR_TWO = '"%s/%s"'

FILE_NAME_SEPARATOR = "."
ANY_NAME = "*"
USB_NAME_RE = "^usb\d$"
GLOBAL = "?global"

RW = "rw"
RO = "ro"


class PalNetatalk(palcmd.PalCmd):
    """Pal Netatalk command implementation class."""

    def __check_root_node_name(self, entry, name):
        """Root node name check."""

        if entry.get().get_node_name() != name:
            raise InvalidArgError(
                    "root_node_name = " + entry.get().get_node_name())

    def __is_enabled(self, node, path):
        """Check enabled. In the case of NotFoundError, return False"""

        try:
            result = node.get_node(path).is_enabled()
            return result

        except NotFoundError:
            return False

    def __get_detail_permission_list(self, entry, format_str, path):
        """Get detail permission list."""

        list = []
        try:
            list = [format_str % node.get_node(NAME_NAME).get_text() \
                    for node in entry.get().get_node_list(path) \
                    if node.get_node(NAME_NAME).get_text() != ""]

        except NotFoundError:
            pass

        return list

    def __is_read_only(self, node, suffix):
        """Is read only check."""

        if suffix == RO:
            return True
        if self.__is_enabled(node, READONLY_PATH):
            return True
        if self.__is_enabled(node, QUOTA_PATH):
            return True
        return False

    def __get_detail_permission_share_setting(self, entry):
        """Get detail permission share setting."""

        value_dict = {}
        if self.__is_enabled(entry.get(), DETAILPERMISSION_PATH) == False:
            value_dict[ALLOW_KEY] = "nobody,@nogroup"
            if (self.__is_enabled(entry.get(), READONLY_PATH) or
                    self.__is_enabled(entry.get(), QUOTA_PATH)):
                return value_dict
            value_dict[RWLIST_KEY] = "nobody,@nogroup"
            return value_dict

        rou_list = rog_list = rwu_list = rwg_list = []

        rou_list = self.__get_detail_permission_list(entry,
                USER_FMT_STR, RO_USER_LIST_PATH)
        rog_list = self.__get_detail_permission_list(entry,
                GROUP_FMT_STR, RO_GROUP_LIST_PATH)
        rwu_list = self.__get_detail_permission_list(entry,
                USER_FMT_STR, RW_USER_LIST_PATH)
        rwg_list = self.__get_detail_permission_list(entry,
                GROUP_FMT_STR, RW_GROUP_LIST_PATH)

        ug_valid = \
            ",".join(["admin"] + rou_list + rwu_list + rog_list + rwg_list)
        value_dict[ALLOW_KEY] = ug_valid
        value_dict[ROLIST_KEY] = ",".join(rou_list + rog_list)
        value_dict[RWLIST_KEY] = ",".join(["admin"] + rwu_list + rwg_list)

        return value_dict

    def __get_options_set(self, suffix, entry):
        """Make a value_dict for every temp conf file."""

        value_dict = {}
        if self.__is_read_only(entry.get(), suffix):
            if self.__timemachine_is_enable(entry):
                value_dict[OPTIONS_KEY] = "ro,tm"
            else:
                value_dict[OPTIONS_KEY] = "ro"
        else:
            if self.__timemachine_is_enable(entry):
                value_dict[OPTIONS_KEY] = "tm"

        return value_dict

    def __create_applevolumes_default_part_file(self, entry, file_dir):
        """Create AppleVolumes.default file."""

        value_dict = {}
        share_name = entry.get().get_node(NAME_PATH).get_text()
        share_path = entry.get().get_node(PATH_PATH).get_text()
        value_dict[DBPATH_KEY] = \
            AFP_CONF_ITEM_FMT_STR_TWO % (config.APPLEDB_DIR, share_name)

        for suffix in [RW, RO]:
            value_dict.update(self.__get_options_set(suffix, entry))
            value_dict.update(
                    self.__get_detail_permission_share_setting(entry))

            tmp_file_name = share_name + FILE_NAME_SEPARATOR + \
                    suffix + FILE_NAME_SEPARATOR + str(os.getpid())
            tmp_file_path = os.path.join(config.TEMP_DIR, tmp_file_name)

            tmp_file = open(tmp_file_path, WRITE_MODE)
            tmp_file.write(AFP_CONF_COMMENT_FMT_STR % share_name)
            tmp_file.write(AFP_CONF_ITEM_FMT_STR_ONE %
                    (share_path, share_name))

            for key in (OPTIONS_KEY, DBPATH_KEY, ALLOW_KEY,
                        RWLIST_KEY, ROLIST_KEY):
                if key in value_dict:
                    tmp_file.write(" " + key + ":" + value_dict[key])
            tmp_file.write("\n")
            tmp_file.close()

            file_name = share_name + FILE_NAME_SEPARATOR + suffix
            link_path = ""
            usb_match = re.match(USB_NAME_RE, share_name)
            if usb_match:
                link_path = os.path.join(file_dir, file_name)
                file_name += USB_ORIG_SUFFIX 
            file_path = os.path.join(file_dir, file_name)
            shutil.move(tmp_file_path, file_path)
            if link_path != "":
                if not os.path.exists(link_path):
                    os.symlink(file_path, link_path)

        return

    def __modify_timemachine_service_add(self, entry):
        """Modify timemachine.service file when a share folder is added."""

        (share_name, conf_path, tmp_file_path) = self.__get_basic_info(entry)

        if (os.path.exists(conf_path) == False or
                            len(open(conf_path, READ_MODE).read()) == 0):
            tmp_file = open(tmp_file_path, WRITE_MODE)
            tmp_file.write(TIMEMACHINE_SERVICE_FMT_STR % share_name)
            tmp_file.close()
        else:
            match_str = "adVN=%s," % share_name
            if match_str in open(conf_path, READ_MODE).read():
                return

            conf_data = xmldata.XmlData(conf_path)
            service_node = conf_data.get_root_node().get_node(SERVICE)
            txt_list = service_node.get_node_list(TXTRECORD)
            new_node = xmldata.XmlNode(TXTRECORD)
            new_node_text = TXTRECORD_FMT_STR % (str(len(txt_list) - 1),
                                                            share_name)
            new_node.set_text(new_node_text)
            service_node.append_child(new_node)
            conf_data.save(tmp_file_path)

        self.__make_timemachine_config(conf_path, tmp_file_path)
        return

    def __make_timemachine_config(self, conf_path, tmp_file_path):

        head_file_path = os.path.join(config.TEMP_DIR, "timemachine.head")
        head_file = open(head_file_path, WRITE_MODE)
        head_file.write(TIMEMACHINE_SERVICE_STATE)
        head_file.close()

        build_config_file(conf_path, [head_file_path, tmp_file_path])
        os.remove(tmp_file_path)
        os.remove(head_file_path)

        return

    def __get_basic_info(self, entry):

        dir_path = config.AVAHI_SERVICES_REAL_DIR
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        conf_path = os.path.join(dir_path, config.TIMEMACHINE_SERVICE_FILE)
        share_name = entry.get().get_node(NAME_PATH).get_text()
        tmp_file_name = TIMEMACHINE + FILE_NAME_SEPARATOR + str(os.getpid())
        tmp_file_path = os.path.join(config.TEMP_DIR, tmp_file_name)

        return (share_name, conf_path, tmp_file_path)

    def __create_share_conf_part_file(self, entry, src_dir):
        """Create netatalk share part file."""

        if not self.__is_enabled(entry.get(), APPLESHARE_PATH):
            return

        if not os.path.exists(src_dir):
            os.makedirs(src_dir)

        self.__create_applevolumes_default_part_file(entry, src_dir)

        return

    def __delete_share_conf_part_file(self, entry, src_dir):
        """Delete netatalk share part file."""

        share_name = entry.get().get_node(NAME_PATH).get_text()
        for suffix in [RW, RO]:
            del_list = []
            del_list.append(share_name + FILE_NAME_SEPARATOR + suffix)

            usb_match = re.match(USB_NAME_RE, share_name)
            if usb_match:
                file_name = share_name + FILE_NAME_SEPARATOR + \
                            suffix + USB_ORIG_SUFFIX
                del_list.append(file_name)

            for name in del_list:
                file_path = os.path.join(src_dir, name)
                if os.path.exists(file_path):
                    os.remove(file_path)
        return

    def __modify_timemachine_service_delete(self, entry):
        """Modify timemachine.service file when a share folder is deleted."""

        (share_name, conf_path, tmp_file_path) = self.__get_basic_info(entry)
        if not os.path.exists(conf_path):
            return
        match_str1 = "adVN=%s," % share_name
        if match_str1 in open(conf_path, READ_MODE).read():
            data = xmldata.XmlData(conf_path)

            service = data.get_node(SERVICE)
            num = 0
            for child in service.get_node_list(TXTRECORD):
                if match_str1 in child.get_text():
                    break
                num += 1
            service.delete_child(TXTRECORD, index=num)

            num = 0
            for child in service.get_node_list(TXTRECORD):
                (child_text, str_num) = re.subn(r"^dk([0-9]+)=",
                                  "dk%s=" % num, child.get_text())
                if str_num == 1:
                    child.set_text(child_text)
                    num += 1
            data.save(tmp_file_path)
            self.__make_timemachine_config(conf_path, tmp_file_path)
        return

    def __get_mount_info(self, vol):
        """Get the suffix list of 'vol'."""

        mnt_point = config.VOL_INFO[vol]["mnt_point"]
        if util.is_mounted(mnt_point):
            if util.is_mounted_as_read_only(mnt_point):
                return [".ro"]
            else:
                return [".rw"]
        else:
            return None

    def __sort_dir_list(self, setting_dir):

        path_list = []
        path_list.append(os.path.join(setting_dir, GLOBAL))

        all_path = glob.glob(os.path.join(setting_dir, ANY_NAME))
        all_path.sort()
        for pt in all_path:
            base = os.path.basename(pt)
            if base != GLOBAL:
                path_list.append(pt)

        return path_list

    def __make_global_part(self, src_dir):

        global_path = os.path.join(src_dir, GLOBAL)
        if not os.path.exists(global_path):
            tmp_global_file = GLOBAL + FILE_NAME_SEPARATOR + str(os.getpid())
            tmp_global_path = os.path.join(config.TEMP_DIR, tmp_global_file)

            file = open(tmp_global_path, WRITE_MODE)
            file.write(GLOBAL_STR)
            file.close()
            shutil.move(tmp_global_path, global_path)

        return

    def __create_applevolumes_default_conf_file(self):
        """Create Applevolumes.default file."""

        config_path = os.path.join(config.AFP_CONF_REAL_DIR,
                                config.APPLEVOLUMES_DEFAULT_FILE)
        setting_dir = config.NETATALK_SETTING_DIR
        try:
            shutil.rmtree(setting_dir)
        except:
            pass
        os.makedirs(setting_dir)

        src_dir = config.AFP_CONF_PART_DIR
        self.__make_global_part(src_dir)
        enable_setting([GLOBAL], [], src_dir, setting_dir)
        self.__create_or_delete_share_link(src_dir, setting_dir)
        build_config_file(config_path, self.__sort_dir_list(setting_dir))

        return

    def __create_or_delete_share_link(self, src_dir, setting_dir):
        for vol in config.VOL_INFO.keys():
            vol_dir = os.path.join(src_dir, vol)
            share_list = []
            for path in glob.glob(os.path.join(vol_dir, ANY_NAME)):
                base = os.path.basename(path)
                match = re.match("(.*)\.r[ow]$", base)
                if not match:
                    continue
                if match.group(1) in share_list:
                    continue
                share_list.append(match.group(1))
            mnt_opt = self.__get_mount_info(vol)
            if share_list != []:
                if mnt_opt:
                    enable_setting(share_list, mnt_opt, vol_dir, setting_dir)
                else:
                    disable_setting(share_list, setting_dir)

        return

    def __get_src_path(self, entry):
        share_volume = re.match(VOLUME_RE,
                 entry.get().get_node(PATH_PATH).get_text()).group(1)
        return os.path.join(config.AFP_CONF_PART_DIR, share_volume)

    def __timemachine_is_enable(self, entry):
        try:
            tm_node = entry.get().get_node(APPLESHARE_PATH)
            if tm_node.get_attr_value("timemachine") == "true":
                return True
            else:
                return False
        except NotFoundError:
            return False

    def __add_handler(self, entry):
        """Add Handler."""

        try:
            try:
                entry.get().get_node('share/shareservice/appleshare')
                flag = True
            except:
                flag = False
            if flag:
                self.__check_root_node_name(entry, SHARES_NAME)

                src_dir = self.__get_src_path(entry)
                self.__create_share_conf_part_file(entry, src_dir)
                self.__create_applevolumes_default_conf_file()

                if self.__timemachine_is_enable(entry):
                    self.__modify_timemachine_service_add(entry)
            entry.set_result(Results.OK)

        except (IOError, OSError):
            entry.set_result(Results.ADD_ERROR)
            self.exception("Add fail!!")

        return entry

    def __delete_handler(self, entry):
        """Delete Handler."""

        try:
            self.__check_root_node_name(entry, SHARES_NAME)

            src_dir = self.__get_src_path(entry)
            self.__delete_share_conf_part_file(entry, src_dir)
            self.__create_applevolumes_default_conf_file()

            self.__modify_timemachine_service_delete(entry)
            entry.set_result(Results.OK)

        except (IOError, OSError):
            entry.set_result(Results.DELETE_ERROR)
            self.exception("Delete fail!!")

        return entry

    def __modify_handler(self, entry):
        """Modify Handler."""

        try:
            self.__check_root_node_name(entry, SHARES_NAME)

            src_dir = self.__get_src_path(entry)
            self.__delete_share_conf_part_file(entry, src_dir)
            self.__create_share_conf_part_file(entry, src_dir)
            self.__create_applevolumes_default_conf_file()

            if self.__timemachine_is_enable(entry):
                self.__modify_timemachine_service_add(entry)
            else:
                self.__modify_timemachine_service_delete(entry)
            entry.set_result(Results.OK)

        except (IOError, OSError):
            entry.set_result(Results.MODIFY_ERROR)
            self.exception("Modify fail!!")

        return entry

    def __update_handler(self, entry):
        """Update Handler."""

        try:
            self.__create_applevolumes_default_conf_file()
            entry.set_result(Results.OK)

        except (IOError, OSError):
            entry.set_result(Results.UPDATE_ERROR)
            self.exception("Update fail!!")

        return entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.ADD, self.__add_handler)
        self.set_handler(ReqTypes.DELETE, self.__delete_handler)
        self.set_handler(ReqTypes.MODIFY, self.__modify_handler)
        self.set_handler(ReqTypes.UPDATE, self.__update_handler)


#
# Main
#

if __name__ == "__main__":
    PalNetatalk(option=palcmd.PalCmdOption()).handle_request()
