#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for power setting."""

import os

from iobase.cmdexec import (CmdExec, NULL)
from iobase.exception import (NotFoundError, ExecError, InvalidArgError)

import libpal.palcmd as palcmd
from pal_info import (ReqTypes, Results)

import pal_config as config

from libpal.crond import ctrl as crond


POWER_NAME = "powersetting"
AUTORESUME_NAME = "autoresume"
ONTIME_NAME = "ontime"
OFFTIME_NAME = "offtime"
DAYSCHEDULE_PATH = "powerschedule/dayschedule"

AUTOBOOT_ENABLE_ARGS = [config.I2CSET_CMD, "-y", "0", "0x50", "0x1d", "0x01"]
AUTOBOOT_DISABLE_ARGS = [config.I2CSET_CMD, "-y", "0", "0x50", "0x1d", "0x00"]

SHUTDOWN_ARGS = [config.POWER_SHUTDOWN_CMD, '-d', '30']
REBOOT_ARGS = [config.POWER_SHUTDOWN_CMD, "-r", '-d', '30']

INVALID_ROOT_TAG_MSG = "Invalid tag name!!"
AUTOBOOT_FAILED_MSG = "Fail applying autoboot setting!!"
RTCCONF_FAILED_MSG = "Fail applying rtc configuration!!"
SHUTDOWN_FAILED_MSG = "Fail shutdowning!!"
REBOOT_FAILED_MSG = "Fail rebooting!!"

DAYS = ('Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat')
POWER_SHUTDOWN_CONF = os.path.join(config.CONF_DIR, 'power', 'shutdown.conf')


class PalPower(palcmd.PalCmd):
    """PAL Power command implementation class."""

    def __check_root_tag(self, entry):
        """Check root tag."""

        root_name = entry.get().get_node_name()

        if root_name != POWER_NAME:
            raise InvalidArgError(INVALID_ROOT_TAG_MSG)

    def __apply_autoboot(self, entry):
        """Handle apply autoboot."""

        try:
            autoresume_node = entry.get().get_node(AUTORESUME_NAME)

            if autoresume_node.is_enabled():
                cmd_args = AUTOBOOT_ENABLE_ARGS
            else:
                cmd_args = AUTOBOOT_DISABLE_ARGS

            CmdExec(cmd_args).execute(stdout=NULL, stderr=NULL)
        except NotFoundError:
            pass

    def __apply_rtcconf(self, entry):
        """Handle apply rtc config."""

        FMT_ENTRYNAME = "shutdown_%s"
        dayschedule_list = entry.get().get_node_list(DAYSCHEDULE_PATH)
        with open(POWER_SHUTDOWN_CONF, 'w') as f:
            for day in DAYS:
                try:
                    crond.delete_entry(FMT_ENTRYNAME % day)
                except OSError:
                    pass

            for dayschedule in dayschedule_list:
                day = dayschedule.get_attr_value('day')
                if not day in DAYS:
                    raise InvalidArgError  # FIXME Error

                for ONOFFTIME_NAME in (ONTIME_NAME, OFFTIME_NAME):
                    try:
                        onofftime = dayschedule.get_node(ONOFFTIME_NAME)
                    except NotFoundError:
                        onofftime = None
                    if onofftime and onofftime.is_enabled():
                        try:
                            hour = int(onofftime.get_node('hour').get_text())
                            min_ = int(onofftime.get_node('minute').get_text())
                        except NotFoundError:
                            raise
                        except ValueError:
                            raise
                        if ONOFFTIME_NAME is ONTIME_NAME:
                            f.write("%s\t%s:%s\n" % (day, hour, min_))
                        elif ONOFFTIME_NAME is OFFTIME_NAME:
                            crond.add_entry(
                                FMT_ENTRYNAME % day,
                                ' '.join(SHUTDOWN_ARGS),
                                user='root',  # FIXME
                                week=str(dict(zip(DAYS, range(0, 7)))[day]),
                                hour=str(hour), minutes=str(min_))

    def __apply_handler(self, entry):
        """Handle apply request."""

        self.__check_root_tag(entry)

        try:
            self.__apply_autoboot(entry)
        except ExecError:
            self.exception(AUTOBOOT_FAILED_MSG)
            entry.set_result(Results.AUTOBOOT_ERROR)
            return entry

        try:
            self.__apply_rtcconf(entry)
        except ExecError:
            self.exception(RTCCONF_FAILED_MSG)
            entry.set_result(Results.RTCCONF_ERROR)
            return entry

        entry.set_result(Results.OK)
        return entry

    def __shutdown_handler(self, entry):
        """Handle shutdown request."""

        try:
            try:
                force = entry.get().get_node('option').is_set_to_true('force')
            except:
                force = False

            cmd = SHUTDOWN_ARGS
            if force:
                cmd = cmd + ['-f']

            CmdExec(cmd).execute(
                    stdout=NULL, stderr=NULL, wait=False, close_fds=True)
            entry.set_result(Results.OK)
        except ExecError:
            self.exception(SHUTDOWN_FAILED_MSG)
            entry.set_result(Results.SHUTDOWN_ERROR)

        return entry

    def __reboot_handler(self, entry):
        """Handle reboot request."""

        try:
            try:
                force = entry.get().get_node('option').is_set_to_true('force')
            except:
                force = False

            cmd = REBOOT_ARGS
            if force:
                cmd = cmd + ['-f']

            CmdExec(cmd).execute(
                    stdout=NULL, stderr=NULL, wait=False, close_fds=True)
            entry.set_result(Results.OK)
        except ExecError:
            self.exception(REBOOT_FAILED_MSG)
            entry.set_result(Results.REBOOT_ERROR)

        return entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.APPLY, self.__apply_handler)
        self.set_handler(ReqTypes.SHUTDOWN, self.__shutdown_handler)
        self.set_handler(ReqTypes.REBOOT, self.__reboot_handler)

#
# Main
#

if __name__ == "__main__":
    PalPower(option=palcmd.PalCmdOption()).handle_request()
