#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for ProFTPD configuration."""

import os
import glob
import shutil
import re

from iobase.exception import (NotFoundError, InvalidArgError)
from ionas.confut import (build_config_file, disable_setting, enable_setting)

from libpal import (palcmd, util)
from pal_info import (ReqTypes, Results)

import pal_config as config


USB_ORIG_SUFFIX = ".real"
# See "evh_ports"
USB_SHARE_RE = "^usb\d$"

ANY_NAME = "*"

FTP_NAME = "ftp"
NUMBER_NAME = "number"
ENCODE_NAME = "encode"
SHARE_NAME = "share"
NAME_NAME = "name"
DESC_NAME = "desc"
PATH_NAME = "path"
QUOTA_NAME = "quota"
LIMIT_NAME = "limit"
DETAILPERMISSION_PATH = \
    "permission/detailpermission"
READONLY_USERLIST_PATH = \
    "permission/detailpermission/readonlymember/userlist/user"
READWRITE_USERLIST_PATH = \
    "permission/detailpermission/readwritemember/userlist/user"
READONLY_GROUPLIST_PATH = \
    "permission/detailpermission/readonlymember/grouplist/group"
READWRITE_GROUPLIST_PATH = \
    "permission/detailpermission/readwritemember/grouplist/group"
READONLY_PATH = "permission/readonly"
MSSHARE_PATH = "shareservice/msshare"
FTP_PATH = "shareservice/ftp"
NOGUEST_PATH = "shareservice/noguest"

APPLY_FAILED_MSG = "Fail applying ProFTPD configuration!!"
ADD_FAILED_MSG = "Fail adding!!"
DELETE_FAILED_MSG = "Fail deleting!!"
PORT_EMPTY_MSG = "Invalid port (empty)!!"
INVALID_ROOT_TAG_MSG_FMT_STR = "Invalid tag name: %s!!"
INVALID_ENCODE_MSG_FMT_STR = "Invalid encode: %s!!"

PID_SURFIX = "." + str(os.getpid())

WRITE_MODE = "w"

UTF_8 = "utf-8"
SHIFT_JIS = "shift-jis"
CP932 = "CP932"

DIRECTIVE_PORT = "Port"
DIRECTIVE_CHARSETREMOTE = "CharsetRemote"

DIRECTIVE_FMT_STR = "%s %s\n"

ALLOW_ALL = "       AllowAll\n"
DENY_ALL = "       DenyAll\n"
ALLOW_USER_FMT_STR = "       AllowUser %s\n"
ALLOW_GROUP_FMT_STR = "       AllowGroup OR %s\n"
HOSTNAME_STR = "ServerName\t%s\n"

VOLUME_RE = r"/mnt/([^/]+)"

RO = "ro"
RW = "rw"


class PalProFTPD(palcmd.PalCmd):
    """PAL ProFTPD command implementation class."""

    def __check_root_tag(self, entry):
        """Check root tag."""

        root_name = entry.get().get_node_name()

        if root_name != FTP_NAME:
            error_info = INVALID_ROOT_TAG_MSG_FMT_STR % root_name
            self.error(error_info)
            raise InvalidArgError(error_info)

    def __get_encode(self, node):
        """Get encode."""

        encode = node.get_text()

        # FIXME 上位から大文字でくれば大文字定数のみでよい
        # SHIFT_JISも上位からCP932ではだめか?
        if encode == UTF_8:
            return UTF_8.upper()
        elif encode == SHIFT_JIS:
            return CP932
        else:
            error_info = INVALID_ENCODE_MSG_FMT_STR % encode
            self.error(error_info)
            raise InvalidArgError(error_info)

    def __get_port(self, node):
        """Get port."""

        # FIXME portが空の場合のチェック必要か?
        port = node.get_node(NUMBER_NAME).get_text()

        if port == "":
            self.error(PORT_EMPTY_MSG)
            raise InvalidArgError(PORT_EMPTY_MSG)
        else:
            return port

    def __make_global_valiable_file(self, entry):
        """Make global valiable file."""

        port = self.__get_port(entry.get().get_node("port[@type='control']"))
        encode = self.__get_encode(entry.get().get_node(ENCODE_NAME))
        global_valiable_data = \
                DIRECTIVE_FMT_STR % (DIRECTIVE_PORT, port) + \
                DIRECTIVE_FMT_STR % (DIRECTIVE_CHARSETREMOTE, encode)

        file_name = config.PROFTPD_PART_GLOBAL_VALIABLE_FILE
        tmp_file_name = file_name + PID_SURFIX
        tmp_file_path = os.path.join(config.TEMP_DIR, tmp_file_name)
        tmp_file = open(tmp_file_path, WRITE_MODE)
        tmp_file.write(global_valiable_data)
        tmp_file.close()

        file_dir = config.PROFTPD_PART_DIR
        if not os.path.exists(file_dir):
            os.makedirs(file_dir)
        dest_path = os.path.join(file_dir, file_name)

        shutil.move(tmp_file_path, dest_path)
        return

    def __get_mount_info(self, vol):
        """Get the suffix list of 'vol'."""

        mnt_point = config.VOL_INFO[vol]["mnt_point"]
        if util.is_mounted(mnt_point):
            if util.is_mounted_as_read_only(mnt_point):
                return [".ro"]
            else:
                return [".rw"]
        else:
            return None

    def __sort_dir_list_proftpd_anon(self, setting_dir):

        path_list = []
        for part_name in [config.PROFTPD_PART_FOR_UNSPECIFIED_FILE]:
            path_list.append(os.path.join(setting_dir, part_name))
        all_path = glob.glob(os.path.join(setting_dir, ANY_NAME))
        all_path.sort()
        for pt in all_path:
            if not pt in path_list:
                path_list.append(pt)
        return path_list

    def __sort_dir_list_proftpd(self, setting_dir):

        path_list = []
        for part_name in [config.PROFTPD_PART_GLOBAL_BASE_FILE, config.PROFTPD_PART_GLOBAL_HOSTNAME_FILE,
                config.PROFTPD_PART_GLOBAL_VALIABLE_FILE, config.PROFTPD_PART_ANONYMOUS_FILE,
                config.PROFTPD_PART_FOR_UNSPECIFIED_FILE]:
            path_list.append(os.path.join(setting_dir, part_name))
        all_path = glob.glob(os.path.join(setting_dir, ANY_NAME))
        all_path.sort()
        for pt in all_path:
            if not pt in path_list:
                path_list.append(pt)
        return path_list

    def __create_or_delete_share_link(self, src_dir, vol, setting_dir):

        vol_dir = os.path.join(src_dir, vol)
        share_list = []
        for path in glob.glob(os.path.join(vol_dir, ANY_NAME)):
            base = os.path.basename(path)
            match = re.match("(.*)\.r[ow]$", base)
            if not match:
                continue
            if match.group(1) in share_list:
                continue
            share_list.append(match.group(1))
        mnt_opt = self.__get_mount_info(vol)
        if share_list != []:
            if mnt_opt:
                enable_setting(share_list, mnt_opt, vol_dir, setting_dir)
            else:
                disable_setting(share_list, setting_dir)

        return

    def _create_or_delete_share_link(self, src_dir, setting_dir):

        for vol in config.VOL_INFO.keys():
            self.__create_or_delete_share_link(
                src_dir, vol, setting_dir)

    def __make_proftpd_conf_file(self):
        """Make proftpd conf file."""

        conf_dir = config.PROFTPD_REAL_DIR
        if not os.path.exists(conf_dir):
            os.makedirs(conf_dir)
        conf_path = os.path.join(conf_dir, config.PROFTPD_CONF_FILE)
        setting_dir = config.PROFTPD_SETTING_DIR
        if os.path.exists(setting_dir):
            for file in glob.glob(os.path.join(setting_dir, ANY_NAME)):
                if os.path.islink(file):
                    os.remove(file)
        else:
            os.makedirs(setting_dir)

        src_dir = config.PROFTPD_PART_DIR
        enable_setting([config.PROFTPD_PART_GLOBAL_BASE_FILE,
                        config.PROFTPD_PART_GLOBAL_HOSTNAME_FILE,
                        config.PROFTPD_PART_GLOBAL_VALIABLE_FILE,
                        config.PROFTPD_PART_ANONYMOUS_FILE,
                        config.PROFTPD_PART_FOR_UNSPECIFIED_FILE],
                        [], src_dir, setting_dir)
        self._create_or_delete_share_link(src_dir, setting_dir)
        build_config_file(conf_path, self.__sort_dir_list_proftpd(setting_dir))

        return

    def __make_proftpd_anon_conf_file(self):
        """Make proftpd_anon conf file."""

        conf_dir = config.PROFTPD_REAL_DIR
        if not os.path.exists(conf_dir):
            os.makedirs(conf_dir)
        conf_path = os.path.join(conf_dir, config.PROFTPD_ANON_CONF_FILE)
        setting_dir = config.PROFTPD_ANON_SETTING_DIR
        if os.path.exists(setting_dir):
            for file in glob.glob(os.path.join(setting_dir, ANY_NAME)):
                if os.path.islink(file):
                    os.remove(file)
        else:
            os.makedirs(setting_dir)

        src_dir = config.PROFTPD_ANON_PART_DIR
        enable_setting([config.PROFTPD_PART_FOR_UNSPECIFIED_FILE],
                        [], src_dir, setting_dir)
        self._create_or_delete_share_link(src_dir, setting_dir)
        build_config_file(conf_path,
                self.__sort_dir_list_proftpd_anon(setting_dir))

        return

    def __get_name_list(self, share_node, path):
        """Get name list."""

        name_list = []
        try:
            name_list = [node.get_node(NAME_NAME).get_text() \
                    for node in share_node.get_node_list(path)]
        except NotFoundError:
            pass

        return name_list

    def __get_allow_list_string(self, name_list, fmt_str):
        """Get list string."""

        if name_list:
            name_list_string = fmt_str % ",".join(name_list)
        else:
            name_list_string = ""

        return name_list_string

    def __is_readable(self, share_node, anon):
        try:
            ftp_node = share_node.get_node(FTP_PATH)
        except:
            return False
        if ftp_node.is_enabled():
            if anon:
                if share_node.get_node(NOGUEST_PATH).is_enabled():
                    return False
                if share_node.get_node(DETAILPERMISSION_PATH).is_enabled():
                    return False
                return True
            else:
                if share_node.get_node(DETAILPERMISSION_PATH).is_enabled():
                    return True
                return False
            # FIXME 使用量制限が有効時のパターン抜けている

        else:
            return False

    def __is_quota_enabled(self, share_node):
        """Is quota enabled?"""

        enable = False
        msshare_node = share_node.get_node(MSSHARE_PATH)

        if msshare_node.is_enabled() == True:

            quota_node = msshare_node.get_node(QUOTA_NAME)

            if quota_node.is_enabled() == True:

                limit = int(quota_node.get_node(LIMIT_NAME).get_text())

                if limit > 0:
                    enable = True

        return enable

    def __is_writable(self, share_node, readwrite, anon):
        """Is writable?"""

        try:
            ftp_node = share_node.get_node(FTP_PATH)
        except:
            return False
        if ftp_node.is_enabled() == True and \
            share_node.get_node(READONLY_PATH).is_enabled() != True and \
            self.__is_quota_enabled(share_node) != True and \
            readwrite == True:

            if anon == True:
                if share_node.get_node(NOGUEST_PATH).is_enabled() == True:
                    return False
                if share_node.get_node(DETAILPERMISSION_PATH).is_enabled() \
                                                                     == True:
                    return False
                return True
            else:
                if share_node.get_node(DETAILPERMISSION_PATH).is_enabled() \
                                                                     == True:
                    return True
                return False
        else:
            return False

    def __get_read_user(self, share_node):
        """Get readonly user list."""

        user_list = self.__get_name_list(share_node, READONLY_USERLIST_PATH)
        user_list += self.__get_name_list(share_node, READWRITE_USERLIST_PATH)

        return self.__get_allow_list_string(user_list, ALLOW_USER_FMT_STR)

    def __get_write_user(self, share_node, readwrite):
        """Get readwrite user list."""

        user_list = self.__get_name_list(share_node, READWRITE_USERLIST_PATH)

        return self.__get_allow_list_string(user_list, ALLOW_USER_FMT_STR)

    def __get_read_group(self, share_node):
        """Get readonly group list."""

        group_list = self.__get_name_list(
                share_node, READONLY_GROUPLIST_PATH)
        group_list += self.__get_name_list(
                share_node, READWRITE_GROUPLIST_PATH)

        return self.__get_allow_list_string(group_list, ALLOW_GROUP_FMT_STR)

    def __get_write_group(self, share_node, readwrite):
        """Get readwrite group list."""

        group_list = self.__get_name_list(share_node, READWRITE_GROUPLIST_PATH)

        return self.__get_allow_list_string(group_list, ALLOW_GROUP_FMT_STR)

    def __get_read_limit(self, share_node, anon):
        """Get read limit."""

        if anon == True:
            if self.__is_readable(share_node, anon):
                read_limit = ALLOW_ALL
            else:
                read_limit = DENY_ALL
        else:
            read_limit = DENY_ALL
            if self.__is_readable(share_node, anon):
                read_limit += self.__get_read_user(share_node) + \
                              self.__get_read_group(share_node)

        return read_limit

    def __get_write_limit(self, share_node, readwrite, anon):
        """Get write limit."""

        if anon == True:

            if self.__is_writable(share_node, readwrite, anon) == True:
                write_limit = ALLOW_ALL

            else:
                write_limit = DENY_ALL
        else:
            write_limit = DENY_ALL

            if self.__is_writable(share_node, readwrite, anon) == True:
                write_limit += self.__get_write_user(share_node, readwrite) + \
                               self.__get_write_group(share_node, readwrite)

        return write_limit

    def __get_path(self, share_node):

        before_path = share_node.get_node(PATH_NAME).get_text()
        name = share_node.get_node(NAME_NAME).get_text()

        path = ""
        for vol in config.VOL_INFO.keys():
            if before_path.startswith(config.VOL_INFO[vol]["mnt_point"]):
                path += config.VOL_INFO[vol]["alt_mnt_point"]

                # FIXME share_nameをpathに含めるかの判断方法は
                # これでよいか
                if config.VOL_INFO[vol]["share_root"]:
                    path = os.path.join(path, name)

                break

        return path

    def __get_share_part_file_data(self, share_node, readwrite, anon):
        """Get share part file data."""

        path = self.__get_path(share_node)
        desc = share_node.get_node(DESC_NAME).get_text()

        read_limit = self.__get_read_limit(share_node, anon)
        write_limit = self.__get_write_limit(share_node, readwrite, anon)

        file_data = \
            "#\n" + \
            "# " + desc + "\n" + \
            "#\n" + \
            "<Directory \"" + path + "\">\n" + \
            "    <Limit DIRS READ>\n" + \
                    read_limit + \
            "    </Limit>\n" + \
            "    <Limit WRITE>\n" + \
                    write_limit + \
            "    </Limit>\n" + \
            "    <Limit SITE_CHMOD SITE_CHGRP SITE_QUOTA>\n" + \
            "       DenyAll\n" + \
            "    </Limit>\n" + \
            "</Directory>\n"

        return file_data

    def __get_share_part_dir_path(self, path, anon):
        """Get share part directory path."""

        if anon == True:
            part_dir = config.PROFTPD_ANON_PART_DIR
        else:
            part_dir = config.PROFTPD_PART_DIR

        volume = re.match(VOLUME_RE, path).group(1)

        dir_path = os.path.join(part_dir, volume)

        return dir_path

    def __move_share_part_file(self, temp_file_path, anon, path):
        """Move share part file."""

        link_path = ""
        dir_path = self.__get_share_part_dir_path(path, anon)
        file_name = os.path.splitext(os.path.basename(temp_file_path))[0]
        share_name = os.path.splitext(file_name)[0]
        usb_match = re.match(USB_SHARE_RE, share_name)
        if usb_match:
            link_path = os.path.join(dir_path, file_name)
            file_name += USB_ORIG_SUFFIX
        file_path = os.path.join(dir_path, file_name)

        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        shutil.move(temp_file_path, file_path)
        if link_path != "":
            if not os.path.exists(link_path):
                os.symlink(file_path, link_path)
        return

    def __make_share_part_file(self, share_node, anon):
        """Make share part file."""

        name = share_node.get_node(NAME_NAME).get_text()
        path = share_node.get_node(PATH_NAME).get_text()

        for suffix in [RW, RO]:
            file_name = name + "." + suffix + PID_SURFIX
            tmp_file_path = os.path.join(config.TEMP_DIR, file_name)
            readwrite = False
            if suffix == RW:
                readwrite = True
            data = self.__get_share_part_file_data(share_node, readwrite, anon)
            tmp_file = open(tmp_file_path, WRITE_MODE)
            tmp_file.write(data)
            tmp_file.close()

            self.__move_share_part_file(tmp_file_path, anon, path)
        return

    def __delete_share_part_file(self, share_node, anon):
        """delete share part file."""

        sharename = share_node.get_node(NAME_NAME).get_text()
        path = share_node.get_node(PATH_NAME).get_text()
        for suffix in [RW, RO]:
            del_list = []
            del_list.append(sharename + "." + suffix)

            usb_match = re.match(USB_SHARE_RE, sharename)
            if usb_match:
                file_name = sharename + "." + suffix + USB_ORIG_SUFFIX
                del_list.append(file_name)

            dir_path = self.__get_share_part_dir_path(path, anon)
            for name in del_list:
                file_path = os.path.join(dir_path, name)
                if os.path.exists(file_path):
                    os.remove(file_path)
        return

    def __create_hostname_part_file(self, name):

        tmp_file_name = config.PROFTPD_PART_GLOBAL_HOSTNAME_FILE + PID_SURFIX
        tmp_file_path = os.path.join(config.TEMP_DIR, tmp_file_name)
        tmp_file = open(tmp_file_path, WRITE_MODE)
        tmp_file.write(HOSTNAME_STR % name)
        tmp_file.close()

        file_path = os.path.join(config.PROFTPD_PART_DIR,
                            config.PROFTPD_PART_GLOBAL_HOSTNAME_FILE)
        shutil.move(tmp_file_path, file_path)
        return

    def __set_ftp(self, entry):
        """FTP settings."""

        self.__make_global_valiable_file(entry)

        self.__make_proftpd_conf_file()
        self.__make_proftpd_anon_conf_file()

        return

    def __set_hostname(self, entry):
        """Host name settings."""

        host_name = entry.get().get_node("hostdesc/name").get_text()
        self.__create_hostname_part_file(host_name)
        self.__make_proftpd_conf_file()
        return

    def __apply_handler(self, entry):
        """Handle apply request."""

        try:
            root_name = entry.get().get_node_name()
            if root_name == FTP_NAME:
                self.__set_ftp(entry)
            else:
                if len(entry.get().get_node_list("hostdesc")) == 1:
                    self.__set_hostname(entry)
                else:
                    raise InvalidArgError("No hostdesc tag")

            entry.set_result(Results.OK)

        except (IOError, OSError):

            entry.set_result(Results.APPLY_ERROR)
            self.exception(APPLY_FAILED_MSG)

        return entry

    def __add_handler(self, entry):
        """Handle add request."""

        try:
            try:
                entry.get().get_node('share/shareservice/ftp')
                flag = True
            except:
                flag = False
            if flag:
                self.__check_root_tag(entry)

                self.__make_share_part_file(
                        entry.get().get_node(SHARE_NAME), anon=False)
                self.__make_proftpd_conf_file()

                self.__make_share_part_file(
                        entry.get().get_node(SHARE_NAME), anon=True)
                self.__make_proftpd_anon_conf_file()

            entry.set_result(Results.OK)

        except (IOError, OSError):
            entry.set_result(Results.ADD_ERROR)
            self.exception(ADD_FAILED_MSG)

        return entry

    def __delete_handler(self, entry):
        """Handle delete request."""

        try:
            self.__check_root_tag(entry)

            self.__delete_share_part_file(
                    entry.get().get_node(SHARE_NAME), anon=False)
            self.__make_proftpd_conf_file()

            self.__delete_share_part_file(
                    entry.get().get_node(SHARE_NAME), anon=True)
            self.__make_proftpd_anon_conf_file()

            entry.set_result(Results.OK)

        except (IOError, OSError):
            entry.set_result(Results.DELETE_ERROR)
            self.exception(DELETE_FAILED_MSG)

        return entry

    def __update_handler(self, entry):
        """Update Handler."""

        try:
            self.__make_proftpd_conf_file()
            self.__make_proftpd_anon_conf_file()
            entry.set_result(Results.OK)

        except (IOError, OSError):
            entry.set_result(Results.UPDATE_ERROR)
            self.exception("Update fail!!")

        return entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.APPLY, self.__apply_handler)
        self.set_handler(ReqTypes.ADD, self.__add_handler)
        self.set_handler(ReqTypes.DELETE, self.__delete_handler)
        self.set_handler(ReqTypes.UPDATE, self.__update_handler)

#
# Main
#

if __name__ == "__main__":
    PalProFTPD(option=palcmd.PalCmdOption()).handle_request()
