#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for finding the Proxy Servers."""


import re
import urlparse
import pacparser

import iobase.xmldata as xmldata
import iobase.netut as netut

import libpal.palcmd as palcmd
from pal_info import (ReqTypes, Results)
from iobase.exception import (NotFoundError, InvalidArgError)


PROXY_STR = "PROXY"
DIRECT_STR = "DIRECT"

PROXY_NAME = "proxy"
URL_PATH = "url"

SERVER_LIST_NAME = "serverlist"
SERVER_NAME = "server"
HOST_NAME = "host"
PORT_NAME = "port"

UNDEFINED_RESULT = "undefined"


class __InvalidPacFileError:
    """Invalid PAC File Exception."""

    pass


class PalProxy(palcmd.PalCmd):
    """PAL proxy implementation class."""

    def __get_proxy_url(self, proxy_str):
        """Return the server url in "proxy_str"."""

        if not proxy_str.upper().startswith(PROXY_STR):
            raise NotFoundError("Not starts " + PROXY_STR, proxy_str)

        return proxy_str[len(PROXY_STR):].strip()

    def __get_host_and_port(self, proxy_str):
        """Return the host and port tuple from "server"."""

        return netut.get_host_and_port(self.__get_proxy_url(proxy_str))

    def __get_proxy(self, pacfile, url):
        """Return the proxy server(host and port pair) list."""

        try:
            proxy_list_str = pacparser.just_find_proxy(pacfile, url)
            if proxy_list_str == None or proxy_list_str == UNDEFINED_RESULT:
                raise __InvalidPacFileError
        except:
            raise __InvalidPacFileError

        server_list = []
        for proxy_str in proxy_list_str.split(";"):
            proxy_str = proxy_str.strip()

            if DIRECT_STR == proxy_str.upper():
                server_list.append((DIRECT_STR, ""))
                continue

            try:
                server_list.append(self.__get_host_and_port(proxy_str))
            except:
                self.debug_exception("Fail getting host and port:" + proxy_str)

        return server_list

    def __query_handler(self, entry):
        """Handler for "query" request."""

        proxy_node = entry.get()
        if proxy_node.get_node_name() != PROXY_NAME:
            raise InvalidArgError(proxy_node.get_node_name())

        url = proxy_node.get_node(URL_PATH).get_text()
        if not urlparse.urlsplit(url).scheme:
            raise InvalidArgError("Invalid URL:" + str(url))

        pacfile = entry.get_file()
        server_list_node = xmldata.XmlConfNode(SERVER_LIST_NAME)
        try:
            proxy_server_list = self.__get_proxy(pacfile, url)
            if not len(proxy_server_list):
                entry.set_result(Results.NO_PROXY)
                return entry

            for server in proxy_server_list:
                server_node = xmldata.XmlConfNode(SERVER_NAME)
                server_node.append_child(HOST_NAME, server[0])
                server_node.append_child(PORT_NAME, server[1])
                server_list_node.append_child(server_node)

            proxy_node.append_child(server_list_node)
            entry.set_result(Results.OK)
            return entry

        except:
            entry.set_result(Results.INVALID_FILE)
            return entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.QUERY, self.__query_handler)


#
# Main
#

if __name__ == "__main__":
    PalProxy(option=palcmd.PalCmdOption()).handle_request()
