#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for USER configuration."""

import os
import re

from iobase.cmdexec import (CmdExec, NULL, PIPE)
from iobase.exception import (ExecError, InvalidArgError, NotFoundError)
from iobase.xmldata import (XmlNode)
import libpal.palcmd as palcmd
from pal_info import (ReqTypes, Results)

import pal_config as config


QUOTA_NAME = "quota"

INVALID_ROOT_TAG_MSG = "Invalid tag name!!"

QUOTA_VOLUME_IDX = 3
SETQUOTA_TARGET_IDX = 1
SETQUOTA_NAME_IDX = 2
SETQUOTA_BLOCK_HARD_LIMIT_IDX = 4
SETQUOTA_VOLUME_IDX = 7
REPQUOTA_TARGET_IDX = 1
REPQUOTA_VOLUME_IDX = 3

SETQUOTA_USER_OPT = "-u"
SETQUOTA_GROUP_OPT = "-g"
REPQUOTA_USER_OPT = "-u"
REPQUOTA_GROUP_OPT = "-g"

QUOTAON_CMD_ARGS = [
    config.QUOTAON_CMD,
    "-x", "enforce", ""]
QUOTAOFF_CMD_ARGS = [
    config.QUOTAOFF_CMD,
    "-x", "enforce", ""]
SETQUOTA_CMD_ARGS = [
    config.SETQUOTA_CMD,
    "", "", "0", "", "0", "0", ""]

NAME_NAME = "name"
VOLUME_NAME = "volume"
USERLIST_NAME = "userlist"
GROUPLIST_NAME = "grouplist"
QUOTA_NAME = "quota"
VOLUME_LIST_PATH = "volumelist/*"
USER_LIST_PATH = "userlist/*"
GROUP_LIST_PATH = "grouplist/*"
LIMIT_PATH = "quota/limit"
VOLUME_ATTR = "volume"
RESULT_ATTR = "result"

USER_NODE_FMT_STR = \
    "<user><name>%s</name><quota><used>%s</used>" + \
    "<limit>%s</limit></quota></user>"
GROUP_NODE_FMT_STR = \
    "<group><name>%s</name><quota><used>%s</used>" + \
    "<limit>%s</limit></quota></group>"
SUCCESS_VALUE = "0"
FAIL_VALUE = "1"

PARSE_RE = "[+-][+-]"
BLOCK_USED_IDX = 0
BLOCK_HARD_LIMIT_IDX = 2

QUERY_FAILED_MSG_FMT_STR = "Fail querying!!: %s"

SUCCESS_CMD_RETURN_CODE = 0

ROOT_NAME = "root"
BLOCK_USED_KEY = "bused"
BLOCK_HARD_LIMIT_KEY = "bhlimit"

ALREADY_SET_ERROR = 2


class PalQuota(palcmd.PalCmd):
    """PAL Quota command implementation class."""

    def __check_root_tag(self, entry):
        """Check root tag."""

        root_name = entry.get().get_node_name()

        if root_name != QUOTA_NAME:
            raise InvalidArgError(INVALID_ROOT_TAG_MSG)

    def __pipe_wait(self, proc, cmd_args):
        """Pipe wait."""

        proc.wait()
        if proc.returncode != SUCCESS_CMD_RETURN_CODE:
            raise ExecError(cmd_args)

    def __exec_enable_quota(self, volume_node):

        try:
            volume_node.set_attrs({RESULT_ATTR: FAIL_VALUE})

            if volume_node.is_enabled():
                cmd_args = list(QUOTAON_CMD_ARGS)
            else:
                cmd_args = list(QUOTAOFF_CMD_ARGS)

            cmd_args[QUOTA_VOLUME_IDX] = volume_node.get_text()

            CmdExec(cmd_args).execute(stdout=NULL, stderr=NULL)

            volume_node.set_attrs({RESULT_ATTR: SUCCESS_VALUE})

        except NotFoundError:
            self.exception("Invalid XML")
        except ExecError, ins:
            if ins.returncode != ALREADY_SET_ERROR:
                self.exception("ExecError: " + str(cmd_args))
            volume_node.set_attrs({RESULT_ATTR: SUCCESS_VALUE})

    def __exec_set_quota(self, node, target_opt):

        try:
            node.set_attrs({RESULT_ATTR: FAIL_VALUE})

            name = node.get_node(NAME_NAME).get_text()
            volume = node.get_node(QUOTA_NAME).get_attr_value(VOLUME_ATTR)
            block_hard_limit = self.__gb_to_kb(
                node.get_node(LIMIT_PATH).get_text())

            cmd_args = list(SETQUOTA_CMD_ARGS)
            cmd_args[SETQUOTA_TARGET_IDX] = target_opt
            cmd_args[SETQUOTA_NAME_IDX] = name
            cmd_args[SETQUOTA_BLOCK_HARD_LIMIT_IDX] = block_hard_limit
            cmd_args[SETQUOTA_VOLUME_IDX] = volume

            CmdExec(cmd_args).execute(stdout=NULL, stderr=NULL)

            node.set_attrs({RESULT_ATTR: SUCCESS_VALUE})

        except NotFoundError:
            self.exception("Invalid XML")
        except ExecError:
            self.exception("ExecError: " + str(cmd_args))

    def __enable_quota(self, entry):

        try:
            for volume_node in entry.get().get_node_list(VOLUME_LIST_PATH):
                self.__exec_enable_quota(volume_node)
        except NotFoundError:
            pass

    def __set_user_quota(self, entry):

        try:
            for node in entry.get().get_node_list(USER_LIST_PATH):
                self.__exec_set_quota(node, SETQUOTA_USER_OPT)
        except NotFoundError:
            pass

    def __set_group_quota(self, entry):

        try:
            for node in entry.get().get_node_list(GROUP_LIST_PATH):
                self.__exec_set_quota(node, SETQUOTA_GROUP_OPT)
        except NotFoundError:
            pass

    def __get_user_quota(self, cmd_output_dict):

        userlist_node = XmlNode(USERLIST_NAME)
        for name, value in cmd_output_dict.iteritems():
            user_node = XmlNode(USER_NODE_FMT_STR % (
                name, value[BLOCK_USED_KEY], value[BLOCK_HARD_LIMIT_KEY]))
            userlist_node.append_child(user_node)

        return userlist_node

    def __get_group_quota(self, cmd_output_dict):

        grouplist_node = XmlNode(GROUPLIST_NAME)
        for name, value in cmd_output_dict.iteritems():
            group_node = XmlNode(GROUP_NODE_FMT_STR % (
                name, value[BLOCK_USED_KEY], value[BLOCK_HARD_LIMIT_KEY]))
            grouplist_node.append_child(group_node)

        return grouplist_node

    def __kb_to_gb(self, value):
        """ kbyte to gbyte(decimal point first place)"""

        return str(float(int(float(value) / 1024 / 1024 * 10)) / 10)

    def __gb_to_kb(self, value):
        """ gbyte to kbyte(decimal point first place)"""

        return str(int(float(value) * 1024 * 1024))

    def __parse_output(self, proc):

        cmd_output_dict = {}

        for line in proc.stdout.readlines():
            try:
                name, info = re.split(PARSE_RE, line, 1)
                name = name.strip()

                if not name or not info or name == ROOT_NAME:
                    continue

                block_used = self.__kb_to_gb(
                    info.split()[BLOCK_USED_IDX])
                block_hard_limit = self.__kb_to_gb(
                    info.split()[BLOCK_HARD_LIMIT_IDX])

                cmd_output_dict[name] = {
                    BLOCK_USED_KEY: block_used,
                    BLOCK_HARD_LIMIT_KEY: block_hard_limit}

            except ValueError:
                continue

        return cmd_output_dict

    def __get_cmd_output(self, entry, target_opt):

        cmd_args = [config.REPQUOTA_CMD,
                target_opt, '-v', entry.get().get_value(VOLUME_NAME)]
        proc = CmdExec(cmd_args).execute(stdout=PIPE, stderr=NULL)
        cmd_output_dict = self.__parse_output(proc)

        self.__pipe_wait(proc, cmd_args)

        return cmd_output_dict

    def __apply_handler(self, entry):
        """Handle apply request."""

        self.__check_root_tag(entry)

        self.__enable_quota(entry)

        self.__set_user_quota(entry)

        self.__set_group_quota(entry)

        entry.set_result(Results.OK)

        return entry

    def __query_handler(self, entry):
        """Handle query request."""

        try:
            self.__check_root_tag(entry)

            entry.get().append_child(self.__get_user_quota(
                self.__get_cmd_output(entry, REPQUOTA_USER_OPT)))

            entry.get().append_child(self.__get_group_quota(
                self.__get_cmd_output(entry, REPQUOTA_GROUP_OPT)))

            entry.set_result(Results.OK)

        except ExecError, ins:
            entry.set_result(Results.QUERY_ERROR)
            self.exception(QUERY_FAILED_MSG_FMT_STR % str(ins.info))

        return entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.APPLY, self.__apply_handler)
        self.set_handler(ReqTypes.QUERY, self.__query_handler)

#
# Main
#

if __name__ == "__main__":
    PalQuota(option=palcmd.PalCmdOption()).handle_request()
