#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta
import os
import re
import time
import urllib2
from ionas import request
from iobase.exception import (InvalidArgError, NotFoundError, ExecError)
from iobase.cmdexec import (CmdExec, PIPE, NULL)
from libpal import palcmd
import pal_info
import pal_config
from pal_info import (ReqTypes, Results)
import product_conf
from iobase import fileut
from iobase.xmldata import (XmlNode, XmlConfNode)
from libpal.util import RemountRoot

import remotelink3.ioportal as ioportal_mod
import remotelink3.pincode as pincode_mod
from pkgaddon.pal_remotelink3_config import (NET_RAW_DEV, RL3_CONF_FILE,
                                                CONF_DATA_TMPL,)
from pkgaddon.remotelink3_config import (IOPORTAL_BASE_URL,
                                            CENTER_SERVER_BASE_URL,
                                            PINCODE_PATH,
                                            INFOUPDATE_STATUS_PATH,)


REMOTELINK3_NAME = "remotelink3"
TIMELIMITLIST_NAME = "timelimitlist"
CLIENTLIMITLIST_NAME = "clientlimitlist"

PINCODE_NAME = "pincode"
IOPORTAL_NAME = "ioportal"
ACCOUNT_NAME_PATH = "account/name"
ACCOUNT_PASSWORD_PATH = "account/password"
SERIALNUMBER_PATH = "product/serialnumber"

CONNECTION_NAME = "connection"
UPNP_NAME = "upnp"
HTTP_PORT_XPATH = """portlist/port[@type="http"]"""
HTTPS_PORT_XPATH = """portlist/port[@type="https"]"""
REMAP_ATTR = "remap"

QUERY_RESPONSE_XML_TMPL = """<remotelink3>
    <touserver>
        <statusinfo>
            <status>%s</status>
            <code>%s</code>
        </statusinfo>
    </touserver>
</remotelink3>"""

STATUS_SUCCESS = "SUCCESS"
STATUS_ERROR = "ERROR"
DEFAULT_CODE = "000"
CODE_ERROR_MASK = 0x0ff

IP_ADDR_SHOW_BASE_ARGS = [pal_config.IP_CMD, "-o", "-f", "link", "addr",
                            "show", "dev"]

GET_PINCODE_REQ = "get_pincode"
CHANGE_PINCODE_REQ = "change_pincode"
DELETE_PINCODE_REQ = "delete_pincode"
GET_TIMELIMIT_LIST_REQ = "get_timelimit_list"
GET_CLIENTLIMIT_LIST_REQ = "get_clientlimit_list"
ADD_TIMELIMIT_REQ = "add_timelimit"
DELETE_TIMELIMIT_REQ = "delete_timelimit"
ADD_CLIENTLIMIT_REQ = "add_clientlimit"
DELETE_CLIENTLIMIT_REQ = "delete_clientlimit"

RETRY_COUNT = 3
RETRY_PERIOD = 20


class PALRemoteLink3(palcmd.PalCmd):
    """PAL remotelink3 command implementation class."""

    def _check_body_node(self, node, type=None):
        """Check if entry's body node is valid."""

        if node.get_node_name() != REMOTELINK3_NAME:
            raise InvalidArgError(node.get_node_name())

    def _get_mac_address(self):
        """Return LAN1 MAC address."""

        LINK_FAMILY_RE = ".*:\s*.*:.*link/ether\s*([0-9a-fA-F:]+)\s*brd.*$"

        cmd_args = list(IP_ADDR_SHOW_BASE_ARGS)
        cmd_args = cmd_args + [NET_RAW_DEV]
        proc = CmdExec(cmd_args, stdout=PIPE, stderr=NULL).execute()
        info_str = proc.stdout.read()
        if proc.wait() != 0:
            raise ExecError(cmd_args, proc.returncode)

        match = re.match(LINK_FAMILY_RE, info_str)
        if not match:
            raise NotFoundError("no MAC address")

        return match.group(1)

    def _delete_pin_code_in_center_server(self, pincode,
            center_server_base_url):
        """Delete PIN code in Center Server."""

        count = 0
        while True:
            try:
                pincode.delete(center_server_base_url)
                return
            except pincode_mod.RequestError, err:
                if err.get_result() == pincode.RESULT_ALREADY_DELETED:
                    self.warning("PIN code is already deleted.")
                    return
            except:
                pass

            if count >= RETRY_COUNT:
                raise
            count += 1
            self.exception("retry delete pin code in center server")
            time.sleep(RETRY_PERIOD)

    def _create_pin_code_in_center_server(self, pincode,
            center_server_base_url, encoded_key, ioportal_account,
            ioportal_password, serial_number, mac_address):
        """Create PIN code in Center Server.
           If PIN core is already exists, then try deleting it via IOPortal.
        """

        count = 0
        while True:
            try:
                pincode.create(center_server_base_url, encoded_key)
                return
            except pincode_mod.RequestError, err:
                if err.get_result() == pincode.RESULT_ALREADY_REGISTERED:
                    self.warning("PIN code is already exists. Try deleting.")
                    try:
                        ioportal = ioportal_mod.IOPortal(IOPORTAL_BASE_URL,
                                                    ioportal_account,
                                                    ioportal_password)
                        ioportal.delete_pin_code(mac_address,
                                                    serial_number, "")
                        pincode.create(center_server_base_url, encoded_key)
                        return
                    except:
                        raise
            except:
                pass

            if count >= RETRY_COUNT:
                raise
            count += 1
            self.exception("retry create pin code in center server")
            time.sleep(RETRY_PERIOD)

    def __try_apply_for_ioportal(self, entry):
        """Try applying "ioportal" tag in "entry"."""

        try:
            ioportal_node = entry.get().get_node(IOPORTAL_NAME)
        except:
            # No "ioportal" tag.  Do nothing.
            return

        try:
            pincode = pincode_mod.PinCode()
            pincode.load(PINCODE_PATH)
            if not pincode.get():
                return
        except:
            # No PIN Code is registered.  Do nothing.
            return

        account = ioportal_node.get_node(ACCOUNT_NAME_PATH).get_text()
        password = ioportal_node.get_node(ACCOUNT_PASSWORD_PATH).get_text()
        serial_number = ioportal_node.get_node(SERIALNUMBER_PATH).get_text()
        mac_address = self._get_mac_address()

        ioportal = ioportal_mod.IOPortal(IOPORTAL_BASE_URL, account, password)

        try:
            ioportal.regist_pin_code(mac_address, serial_number, pincode.get())
        except ioportal_mod.RequestError, err:
            if err.get_code() == ioportal_mod.ALREADY_REGISTERED_ERROR:
                # This is no problem.
                return
            if err.get_code() == ioportal_mod.REGISTRATION_FAIL_ERROR:
                self.warning("IOPortal: PIN code registration fail.")
                # Ignore this error.
                return
            raise

    def __try_apply_for_connection(self, entry):
        """Try applying "connection" tag in "entry"."""

        try:
            connection = entry.get().get_node(CONNECTION_NAME)
            http_port = int(connection.get_node(HTTP_PORT_XPATH).get_text())
            https_port = int(connection.get_node(HTTPS_PORT_XPATH).get_text())
        except:
            # No connection is specified. Do nothing.
            return

        use_upnp = 0
        remap = False
        upnp_http_port = http_port
        upnp_https_port = https_port
        try:
            upnp = connection.get_node(UPNP_NAME)
            if upnp.is_enabled():
                use_upnp = 1
            if upnp.is_set_to_true(REMAP_ATTR):
                remap = True

            if remap:
                upnp_http_port = int(upnp.get_value(HTTP_PORT_XPATH,
                                                        str(http_port)))
                upnp_https_port = int(upnp.get_value(HTTPS_PORT_XPATH,
                                                        str(https_port)))
        except:
            # Ignore error
            pass

        conf_data = CONF_DATA_TMPL % {
                        "http_port": http_port,
                        "https_port": https_port,
                        "upnp": use_upnp,
                        "upnp_http_port": upnp_http_port,
                        "upnp_https_port": upnp_https_port,
                    }

        tmp_name = RL3_CONF_FILE + "." + str(os.getpid())
        with open(tmp_name, "w") as conf_file:
            conf_file.write(conf_data)

        os.rename(tmp_name, RL3_CONF_FILE)

    def __get_pincode_handler(self, entry):
        """Handle get_pincode request."""

        self._check_body_node(entry.get())

        try:
            pincode = pincode_mod.PinCode()
            pincode.load(PINCODE_PATH)
            if pincode.get():
                entry.get().append_child(PINCODE_NAME, pincode.get())
                entry.set_result(Results.OK)
            else:
                entry.set_result(Results.GET_ERROR)
        except:
            self.exception("Get Error")
            entry.set_result(Results.GET_ERROR)

        return entry

    def __change_pincode_handler(self, entry):
        """Handle change_pincode request."""

        self._check_body_node(entry.get())

        try:
            pincode = pincode_mod.PinCode()
            pincode.load(PINCODE_PATH)
        except:
            self.exception("Get Error")
            entry.set_result(Results.GET_ERROR)
            return entry

        try:
            ioportal_req = entry.get().get_node(IOPORTAL_NAME)
            account = ioportal_req.get_node(ACCOUNT_NAME_PATH).get_text()
            password = ioportal_req.get_node(ACCOUNT_PASSWORD_PATH).get_text()
            serial_number = ioportal_req.get_node(SERIALNUMBER_PATH).get_text()
            mac_address = self._get_mac_address()

            ioportal = ioportal_mod.IOPortal(IOPORTAL_BASE_URL, account,
                                                password)
            if pincode.get():
                ioportal.delete_pin_code(mac_address, serial_number,
                                            pincode.get())
        except ioportal_mod.RequestError, err:
            if err.get_code() == ioportal_mod.ALREADY_DISABLED_ERROR:
                self.warning("Remote Link 3 is already disabled.")
            elif err.get_code() == ioportal_mod.AUTH_ERROR:
                self.warning("IOPortal authentication error.")
                entry.set_result(Results.AUTH_ERROR)
                return entry
            elif err.get_code() == ioportal_mod.INVALID_MAC_SERIAL_ERROR:
                self.warning("IOPortal MAC address/Serial Number error.")
                entry.set_result(Results.INVALID_DATA)
                return entry
            else:
                self.exception("IOPortal.delete_pin_code() fail")
                entry.set_result(Results.DELETE_ERROR)
                return entry
        except urllib2.URLError, urllib2.HTTPError:
                self.exception("IOPortal connect error")
                entry.set_result(Results.CONNECT_ERROR)
                return entry
        except:
            self.exception("IOPortal.delete_pin_code() fail")
            entry.set_result(Results.DELETE_ERROR)
            return entry

        try:
            if pincode.get():
                self._delete_pin_code_in_center_server(pincode,
                                                    CENTER_SERVER_BASE_URL)
        except urllib2.URLError, urllib2.HTTPError:
                self.exception("pincode connect error")
                entry.set_result(Results.CONNECT_ERROR)
                return entry
        except:
            self.exception("PinCode.delete() fail")
            entry.set_result(Results.DELETE_ERROR)
            return entry

        try:
            encoded_key = ioportal.get_encoded_key(mac_address, serial_number)
        except ioportal_mod.RequestError, err:
            if err.get_code() == ioportal_mod.AUTH_ERROR:
                self.warning("IOPortal authentication error.")
                entry.set_result(Results.AUTH_ERROR)
                return entry
            elif err.get_code() == ioportal_mod.INVALID_MAC_SERIAL_ERROR:
                self.warning("IOPortal MAC address/Serial Number error.")
                entry.set_result(Results.INVALID_DATA)
                return entry
            else:
                self.exception("IOPortal.get_encoded_key() fail")
                entry.set_result(Results.CHANGE_ERROR)
                return entry
        except urllib2.URLError, urllib2.HTTPError:
                self.exception("IOPortal connect error")
                entry.set_result(Results.CONNECT_ERROR)
                return entry
        except:
            self.exception("IOPortal.get_encoded_key() fail")
            entry.set_result(Results.CHANGE_ERROR)
            return entry

        try:
            self._create_pin_code_in_center_server(pincode,
                    CENTER_SERVER_BASE_URL, encoded_key, account, password,
                    serial_number, mac_address)
        except urllib2.URLError, urllib2.HTTPError:
                self.exception("pincode connect error")
                entry.set_result(Results.CONNECT_ERROR)
                return entry
        except:
            self.exception("PinCode.create() fail")
            entry.set_result(Results.CHANGE_ERROR)
            return entry

        try:
            ioportal.regist_pin_code(mac_address, serial_number, pincode.get())
        except urllib2.URLError, urllib2.HTTPError:
                self.exception("IOPortal connect error")
                entry.set_result(Results.CONNECT_ERROR)
                return entry
        except:
            self.exception("IOPortal.regist_pin_code() fail")
            entry.set_result(Results.CHANGE_ERROR)
            return entry

        with RemountRoot():
            pincode.save(PINCODE_PATH)
        entry.get().append_child(PINCODE_NAME, pincode.get())
        entry.set_result(Results.OK)
        return entry

    def __delete_pincode_handler(self, entry):
        """Handle delete_pincode request."""

        self._check_body_node(entry.get())

        try:
            pincode = pincode_mod.PinCode()
            pincode.load(PINCODE_PATH)
            if not pincode.get():
                entry.set_result(Results.OK)
                return entry
        except:
            self.exception("Get Error")
            entry.set_result(Results.GET_ERROR)
            return entry

        try:
            ioportal_req = entry.get().get_node(IOPORTAL_NAME)
            account = ioportal_req.get_node(ACCOUNT_NAME_PATH).get_text()
            password = ioportal_req.get_node(ACCOUNT_PASSWORD_PATH).get_text()
            serial_number = ioportal_req.get_node(SERIALNUMBER_PATH).get_text()
            mac_address = self._get_mac_address()

            ioportal = ioportal_mod.IOPortal(IOPORTAL_BASE_URL, account,
                                                password)
            ioportal.delete_pin_code(mac_address, serial_number, pincode.get())
        except ioportal_mod.RequestError, err:
            if err.get_code() == ioportal_mod.ALREADY_DISABLED_ERROR:
                self.warning("Remote Link 3 is already disabled.")
            else:
                self.exception("IOPortal.delete_pin_code() fail")
                entry.set_result(Results.DELETE_ERROR)
                return entry
        except urllib2.URLError, urllib2.HTTPError:
                self.exception("IOPortal connect error")
                entry.set_result(Results.CONNECT_ERROR)
                return entry
        except:
            self.exception("IOPortal.delete_pin_code() fail")
            entry.set_result(Results.DELETE_ERROR)
            return entry

        try:
            self._delete_pin_code_in_center_server(pincode,
                                                    CENTER_SERVER_BASE_URL)
        except urllib2.URLError, urllib2.HTTPError:
                self.exception("pincode connect error")
                entry.set_result(Results.CONNECT_ERROR)
                return entry
        except:
            self.exception("PinCode.delete() fail")
            entry.set_result(Results.DELETE_ERROR)
            return entry

        pincode.set(None)
        with RemountRoot():
            pincode.save(PINCODE_PATH)
        entry.set_result(Results.OK)
        return entry

    def _get_timelimit_list_handler(self, entry):
        """Handle get_timelimit_list request."""

        self._check_body_node(entry.get())
        entry.get().append_child(TIMELIMITLIST_NAME)
        entry.set_result(Results.OK)
        return entry

    def _get_clientlimit_list_handler(self, entry):
        """Handle get_clientlimit_list request."""

        self._check_body_node(entry.get())
        entry.get().append_child(CLIENTLIMITLIST_NAME)
        entry.set_result(Results.OK)
        return entry

    def _add_timelimit_handler(self, entry):
        """Handle add_timelimit request."""

        self._check_body_node(entry.get())
        entry.set_result(Results.OK)
        return entry

    def _delete_timelimit_handler(self, entry):
        """Handle delete_timelimit request."""

        self._check_body_node(entry.get())
        entry.set_result(Results.OK)
        return entry

    def _add_clientlimit_handler(self, entry):
        """Handle add_clientlimit request."""

        self._check_body_node(entry.get())
        entry.set_result(Results.OK)
        return entry

    def _delete_clientlimit_handler(self, entry):
        """Handle delete_clientlimit request."""

        self._check_body_node(entry.get())
        entry.set_result(Results.OK)
        return entry

    def _apply_handler(self, entry):
        """Handle apply request."""

        self._check_body_node(entry.get())
        try:
            self.__try_apply_for_ioportal(entry)
        except ioportal_mod.RequestError, err:
            if err.get_code() == ioportal_mod.AUTH_ERROR:
                self.warning("IOPortal authentication error.")
                entry.set_result(Results.AUTH_ERROR)
                return entry
            elif err.get_code() == ioportal_mod.INVALID_MAC_SERIAL_ERROR:
                self.warning("IOPortal MAC address/Serial Number error.")
                entry.set_result(Results.INVALID_DATA)
                return entry
            else:
                self.exception("IOPortal Request error %d" % (err.get_code(),))
                entry.set_result(Results.APPLY_ERROR)
                return entry
        except urllib2.URLError, urllib2.HTTPError:
            self.exception("IOPortal connect error")
            entry.set_result(Results.CONNECT_ERROR)
            return entry
        except:
            self.exception("Apply IOPortal fail")
            entry.set_result(Results.APPLY_ERROR)
            return entry

        try:
            self.__try_apply_for_connection(entry)
        except:
            self.exception("Apply connection fail")
            entry.set_result(Results.APPLY_ERROR)
            return entry

        entry.set_result(Results.OK)
        return entry

    def _query_handler(self, entry):
        """Handle query request."""

        self._check_body_node(entry.get())

        try:
            with open(INFOUPDATE_STATUS_PATH) as status_file:
                code = status_file.read().strip()
                int(code, 16)
        except:
            # Assume no error status
            code = DEFAULT_CODE

        status = STATUS_SUCCESS
        if int(code, 16) & CODE_ERROR_MASK:
            status = STATUS_ERROR

        response_entry = request.Entry(ReqTypes.QUERY,
                                QUERY_RESPONSE_XML_TMPL % (status, code))
        response_entry.set_result(Results.OK)
        return response_entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(GET_PINCODE_REQ, self.__get_pincode_handler)
        self.set_handler(CHANGE_PINCODE_REQ, self.__change_pincode_handler)
        self.set_handler(DELETE_PINCODE_REQ, self.__delete_pincode_handler)

        self.set_handler(GET_TIMELIMIT_LIST_REQ,
                            self._get_timelimit_list_handler)
        self.set_handler(GET_CLIENTLIMIT_LIST_REQ,
                            self._get_clientlimit_list_handler)
        self.set_handler(ADD_TIMELIMIT_REQ, self._add_timelimit_handler)
        self.set_handler(DELETE_TIMELIMIT_REQ, self._delete_timelimit_handler)
        self.set_handler(ADD_CLIENTLIMIT_REQ, self._add_clientlimit_handler)
        self.set_handler(DELETE_CLIENTLIMIT_REQ,
                            self._delete_clientlimit_handler)

        self.set_handler(ReqTypes.APPLY, self._apply_handler)
        self.set_handler(ReqTypes.QUERY, self._query_handler)


#
# Main
#

if __name__ == "__main__":
    PALRemoteLink3(option=palcmd.PalCmdOption()).handle_request()
