#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for Replication."""

from __future__ import with_statement

import os
import shutil
import errno
import stat
import re

import pal_config as config
from libpal import palcmd, util
from iobase.exception import (InvalidArgError, ExecError)
from iobase.cmdexec import (CmdExec, NULL)
from iobase.xmldata import XmlNode
from ionas.confut import build_config_file, enable_setting, disable_setting
from pal_info import ReqTypes
from pal_info import Results


### XML Path ###
REPLICATION_PATH = "share/replication"
HOST_PATH = REPLICATION_PATH + "/host"
SHARE_PATH = REPLICATION_PATH + "/share"
PASSWD_PATH = REPLICATION_PATH + "/passwd"
USERNAME_PATH = REPLICATION_PATH + "/username"
NAME_PATH = "share/name"
PATH_PATH = "share/path"
UID_PATH = "share/uid"
GID_PATH = "share/gid"
USER_NAME_PATH = "user/name"
USER_PASSWD_PATH = "user/passwd"

### XML Tag Name ###
SHARES_NAME = "shares"
USERS_NAME = "users"
RESULT_NAME = "result"

### Rsync Command ###
RSYNC_ARGS = [config.RSYNC_CMD, "--timeout=60",
    "", ""]
PASSWORD_FILE_OPT_IDX = 2
DST_OPT_IDX = 3

### Fsevsync Command ###
ADDDIR_CMD = "ADDDIR:"
DELDIR_CMD = "DELDIR:"
STOPEQ_CMD = "STOPEQ:"
RESUME_CMD = "RESUME:"

### Module Options in rsyncd.conf ###
MODULE_FMT_STR = \
    "[%s]\n" + \
    "\tpath = %s\n" + \
    "\tuid = %s\n" + \
    "\tgid = %s\n"

### Rsync URL format str ###
RSYNC_URL_FMT_STR = "rsync://%s@[%s]/%s"

### Separator ###
FILE_NAME_SEPARATOR = "."
PASSWORD_FILE_SEPARATOR = ":"

### rsynd.secrets should be 600 Mode ###
PASSWORD_FILE_MODE = (stat.S_IRUSR | stat.S_IWUSR)

### Error Messages ###
CHECK_FAILED_MSG = "Fail checking replication!!"
SUSPEND_FAILED_MSG = "Fail suspend replication!!"
RESUME_FAILED_MSG = "Fail resume replication!!"

### Default ID ###
NOBODY_UID = "65534"
NOGROUP_GID = "65534"

### Volume RE ###
VOLUME_RE = r"/mnt/([^/]+)"

### Share part file suffix ###
# レプリケーション先として公開するのでrwのみつくる
SHARE_PART_FILE_SUFFIX = ".rw"
USB_SHARE_PART_FILE_SUFFIX = ".real"

USB_SHARE_LIST = [
    vol
    for vol
    in config.VOL_INFO.keys()
    if config.VOL_INFO[vol]["share_root"] == ""]

FILE_SEP = "."


class PalReplication(palcmd.PalCmd):
    """Pal Replication implementation class."""

    def __validate_root_tag_name(self, entry, valid_name=SHARES_NAME):
        """Validate the root tag name."""

        root_tag_name = entry.get().get_node_name()
        if valid_name != root_tag_name:
            raise InvalidArgError("invalid root tag..")
        return

    def __set_fsevsync_target(self, entry):

        rep_target_path = entry.get().get_node(PATH_PATH).get_text()
        with open(config.PROC_FSEVSYNC_TARGETS_FILE, "wb") as file:
            file.write(DELDIR_CMD + rep_target_path + "/")
        with open(config.PROC_FSEVSYNC_TARGETS_FILE, "wb") as file:
            file.write(ADDDIR_CMD + rep_target_path + "/")

    def __unset_fsevsync_target(self, entry):

        rep_target_path = entry.get().get_node(PATH_PATH).get_text()
        with open(config.PROC_FSEVSYNC_TARGETS_FILE, "wb") as file:
            file.write(DELDIR_CMD + rep_target_path + "/")

    def __make_info_dir(self, share_name):

        try:
            os.mkdir(config.RSYNC_CLIENT_CONF_DIR)
        except (IOError, OSError), err:
            if err.errno != errno.EEXIST:
                raise
        try:
            os.mkdir(os.path.join(config.RSYNC_CLIENT_CONF_DIR, share_name))
        except (IOError, OSError), err:
            if err.errno != errno.EEXIST:
                raise

    def __set_file(self, share_name, category, text):

        file_path = os.path.join(
            config.RSYNC_CLIENT_CONF_DIR, share_name, category)
        tmp_file_path = os.path.join(
            config.TEMP_DIR, share_name + FILE_SEP + str(os.getpid()))

        with open(tmp_file_path, "wb") as file:
            file.write(text)

        shutil.move(tmp_file_path, file_path)

    def __set_user_name_file(self, share_name, user_name):

        self.__set_file(share_name, "user_name", user_name)

    def __set_password_file(self, share_name, password):

        self.__set_file(share_name, "password", password)

        file_path = os.path.join(
            config.RSYNC_CLIENT_CONF_DIR, share_name, "password")
        os.chmod(file_path, PASSWORD_FILE_MODE)

    def __set_path_file(self, share_name, path):

        self.__set_file(share_name, "path", path)

    def __set_dst_host_file(self, share_name, dst_host):

        self.__set_file(share_name, "dst_host", dst_host)

    def __set_dst_share_file(self, share_name, dst_share):

        self.__set_file(share_name, "dst_share", dst_share)

    def __set_info_files(self, entry):

        body = entry.get()
        share_name = body.get_node(NAME_PATH).get_text()
        share_path = body.get_node(PATH_PATH).get_text()
        dst_user_name = body.get_node(USERNAME_PATH).get_text()
        dst_password = body.get_node(PASSWD_PATH).get_text(do_strip=False)
        dst_host = body.get_node(HOST_PATH).get_text()
        dst_share = body.get_node(SHARE_PATH).get_text()

        self.__make_info_dir(share_name)
        self.__set_user_name_file(share_name, dst_user_name)
        self.__set_password_file(share_name, dst_password)
        self.__set_path_file(share_name, share_path)
        self.__set_dst_host_file(share_name, dst_host)
        self.__set_dst_share_file(share_name, dst_share)

    def __unset_info_files(self, entry):

        share_name = entry.get().get_node(NAME_PATH).get_text()
        share_info_dir = os.path.join(config.RSYNC_CLIENT_CONF_DIR, share_name)

        try:
            shutil.rmtree(share_info_dir)
        except (IOError, OSError), err:
            if err.errno != errno.ENOENT:
                raise

    def __get_conf_part_dir(self, share_path):

        share_volume = self.__get_share_volume(share_path)
        conf_part_dir = os.path.join(
            config.RSYNCD_CONF_PART_DIR, share_volume)

        return conf_part_dir

    def __create_share_part_tmp_file(self,
             share_name, share_path, share_uid, share_gid):

        tmp_file_name = share_name + FILE_NAME_SEPARATOR + str(os.getpid())
        tmp_file_path = os.path.join(config.TEMP_DIR, tmp_file_name)

        with open(tmp_file_path, "wb") as tmp_file:
            tmp_file.write(MODULE_FMT_STR % (
                share_name, share_path, share_uid, share_gid))

        return tmp_file_path

    def __mkdir(self, path):

        try:
            os.makedirs(os.path.dirname(path))
        except (IOError, OSError), err:
            if err.errno != errno.EEXIST:
                raise

    def __get_share_volume(self, share_path):

        return re.match(VOLUME_RE, share_path).group(1)

    def __get_file_path(self, share_path, share_name):

        file_path = os.path.join(
            self.__get_conf_part_dir(share_path), share_name)

        if share_name in USB_SHARE_LIST:
            return file_path + \
                SHARE_PART_FILE_SUFFIX + USB_SHARE_PART_FILE_SUFFIX
        return file_path + SHARE_PART_FILE_SUFFIX

    def __create_share_part_file(self, entry):

        body = entry.get()
        share_name = body.get_value(NAME_PATH)
        share_path = body.get_value(PATH_PATH)
        share_uid = body.get_value(UID_PATH, NOBODY_UID)
        share_gid = body.get_value(GID_PATH, NOGROUP_GID)

        tmp_file_path = self.__create_share_part_tmp_file(
            share_name, share_path, share_uid, share_gid)
        file_path = self.__get_file_path(share_path, share_name)
        self.__mkdir(file_path)
        shutil.move(tmp_file_path, file_path)

    def __delete_share_part_file(self, entry):

        body = entry.get()
        share_name = body.get_node(NAME_PATH).get_text()
        share_path = body.get_node(PATH_PATH).get_text()

        file_path = os.path.join(
            self.__get_conf_part_dir(share_path),
            share_name + SHARE_PART_FILE_SUFFIX)

        try:
            os.remove(file_path)
        except (IOError, OSError), err:
            if err.errno != errno.ENOENT:
                raise

    def __get_part_files_list(self, dir):

        file_list = []
        for dir, dnames, fnames in os.walk(dir):
            file_list += fnames
        return file_list

    def __sort_part_file(self, setting_dir):

        file_list = []
        for file_name in os.listdir(setting_dir):
            file_path = os.path.join(setting_dir, file_name)
            if os.path.basename(file_path) == \
                config.RSYNCD_PART_GLOBAL_BASE_FILE:
                file_list = [file_path] + file_list
            else:
                file_list = file_list + [file_path]
        return file_list

    def __is_rw_mounted(self, vol):

        mnt_point = config.VOL_INFO[vol]["mnt_point"]

        if util.is_mounted(mnt_point):
            if not util.is_mounted_as_read_only(mnt_point):
                return True
        return False

    def __global_conf_setting(self, base_dir, setting_dir):

        GLOBAL_CONF_LIST = [config.RSYNCD_PART_GLOBAL_BASE_FILE]

        enable_setting(GLOBAL_CONF_LIST, [], base_dir, setting_dir)

    def __share_conf_setting(self, base_dir, setting_dir):

        for vol in config.VOL_INFO.keys():
            share_conf_dir = os.path.join(base_dir, vol)
            is_rw_mounted = self.__is_rw_mounted(vol)

            for share_conf in os.listdir(share_conf_dir):
                if not share_conf.endswith(SHARE_PART_FILE_SUFFIX):
                    continue
                if is_rw_mounted:
                    enable_setting([share_conf], [SHARE_PART_FILE_SUFFIX],
                        share_conf_dir, setting_dir)
                else:
                    disable_setting([share_conf], setting_dir)

    def __enable_setting(self, base_dir, setting_dir):

        self.__global_conf_setting(base_dir, setting_dir)
        self.__share_conf_setting(base_dir, setting_dir)

    def __create_rsyncd_conf_file(self):

        config_path = os.path.join(config.RSYNCD_CONF_REAL_DIR,
            config.RSYNCD_CONF_FILE)
        setting_dir = config.RSYNCD_SETTING_DIR

        try:
            shutil.rmtree(setting_dir)
        except:
            pass
        os.makedirs(setting_dir)

        self.__enable_setting(config.RSYNCD_CONF_PART_DIR, setting_dir)

        build_config_file(config_path, self.__sort_part_file(setting_dir))

    def __create_password_tmp_file(self, user_name, password):

        tmp_file_name = config.RSYNCD_PASSWORD_FILE + \
            FILE_NAME_SEPARATOR + str(os.getpid())
        tmp_file_path = os.path.join(config.TEMP_DIR, tmp_file_name)

        with open(tmp_file_path, "wb") as tmp_file:
            tmp_file.write(user_name + PASSWORD_FILE_SEPARATOR + password)

        return tmp_file_path

    def __create_password_file(self, entry):

        user_name = entry.get().get_node(
            USER_NAME_PATH).get_text()
        password = entry.get().get_node(
            USER_PASSWD_PATH).get_text(do_strip=False)

        tmp_file_path = self.__create_password_tmp_file(user_name, password)

        file_path = os.path.join(
            config.RSYNCD_CONF_REAL_DIR, config.RSYNCD_PASSWORD_FILE)

        self.__mkdir(file_path)
        shutil.move(tmp_file_path, file_path)
        os.chmod(file_path, PASSWORD_FILE_MODE)

    def __check(self, entry):

        body = entry.get()
        share_name = body.get_node(NAME_PATH).get_text()
        dst_user_name = body.get_node(USERNAME_PATH).get_text()
        dst_host = body.get_node(HOST_PATH).get_text()
        dst_share = body.get_node(SHARE_PATH).get_text()

        dst_url = RSYNC_URL_FMT_STR % (dst_user_name, dst_host, dst_share)
        password_file_path = os.path.join(
            config.RSYNC_CLIENT_CONF_DIR, share_name, "password")
        password_file_opt = "--password-file=" + password_file_path

        cmd_args = list(RSYNC_ARGS)
        cmd_args[PASSWORD_FILE_OPT_IDX] = password_file_opt
        cmd_args[DST_OPT_IDX] = dst_url

        CmdExec(cmd_args).execute(stdout=NULL, stderr=NULL)

    #
    # Request handlers
    #

    def __set_handler(self, entry):
        """Handle "set" request."""

        self.__validate_root_tag_name(entry)

        self.__set_info_files(entry)
        self.__set_fsevsync_target(entry)

        entry.set_result(Results.OK)
        return entry

    def __unset_handler(self, entry):
        """Handle "unset" request."""

        self.__validate_root_tag_name(entry)

        self.__unset_fsevsync_target(entry)
        self.__unset_info_files(entry)

        entry.set_result(Results.OK)
        return entry

    def __add_handler(self, entry):
        """Handle "add" request."""

        self.__validate_root_tag_name(entry)

        self.__create_share_part_file(entry)
        self.__create_rsyncd_conf_file()

        entry.set_result(Results.OK)
        return entry

    def __delete_handler(self, entry):
        """Handle "delete" request."""

        self.__validate_root_tag_name(entry)

        self.__delete_share_part_file(entry)
        self.__create_rsyncd_conf_file()

        entry.set_result(Results.OK)
        return entry

    def __apply_handler(self, entry):
        """Handle "apply" request."""

        self.__validate_root_tag_name(entry, USERS_NAME)

        self.__create_password_file(entry)

        entry.set_result(Results.OK)
        return entry

    def __check_handler(self, entry):
        """Handle "check" request."""

        self.__validate_root_tag_name(entry, SHARES_NAME)

        try:
            self.__check(entry)
            entry.get().append_child(RESULT_NAME, "0")
            entry.set_result(Results.OK)
        except ExecError, ins:
            entry.get().append_child(RESULT_NAME, str(ins.returncode))
            entry.set_result(Results.CHECK_ERROR)
            self.exception(CHECK_FAILED_MSG)

        return entry

    def __update_handler(self, entry):
        """Handle "update" request."""

        self.__create_rsyncd_conf_file()

        entry.set_result(Results.OK)
        return entry

    def __suspend_handler(self, entry):
        """Handle "update" request."""

        try:
            with open(config.PROC_FSEVSYNC_TARGETS_FILE, "wb") as file:
                file.write(STOPEQ_CMD)
            entry.set_result(Results.OK)
        except:
            entry.set_result(Results.SUSPEND_ERROR)
            self.exception(SUSPEND_FAILED_MSG)
        return entry

    def __resume_handler(self, entry):
        """Handle "update" request."""

        try:
            with open(config.PROC_FSEVSYNC_TARGETS_FILE, "wb") as file:
                file.write(RESUME_CMD)
            entry.set_result(Results.OK)
        except:
            entry.set_result(Results.RESUME_ERROR)
            self.exception(RESUME_FAILED_MSG)
        return entry

    def _report_is_supported(self):
        try:
            with open('/proc/fsevsync/version', 'r') as fh:
                version = fh.read().strip()
                segments = version.split('.')

                _version = '1.01'
                for i, _segment in enumerate(_version.split('.')):
                    if segments[i] < _segment:
                        break
                else:
                    return True
                return False
        except:
            return False

    def __feature_handler(self, entry):
        """Handle "feature" request."""

        try:
            replication_node = entry.get()

            replication_node.delete_children('*')
            feature_node = XmlNode('<feature />')
            for func in ['report']:
                method = getattr(self, "_%s_is_supported" % func)
                try:
                    func_node = XmlNode("""<%s />""" % func)
                    func_node.set_attrs({'enable': str(method()).lower()})
                    feature_node.append_child(func_node)
                except:
                    continue
            replication_node.append_child(feature_node)
        except:
            entry.set_result(Results.INTERNAL_ERROR)
            self.exception('Fail feature replication!!')
        entry.set_result(Results.OK)
        return entry

    #
    # Constractor
    #

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.SET, self.__set_handler)
        self.set_handler(ReqTypes.UNSET, self.__unset_handler)
        self.set_handler(ReqTypes.ADD, self.__add_handler)
        self.set_handler(ReqTypes.DELETE, self.__delete_handler)
        self.set_handler(ReqTypes.APPLY, self.__apply_handler)
        self.set_handler(ReqTypes.CHECK, self.__check_handler)
        self.set_handler(ReqTypes.UPDATE, self.__update_handler)
        self.set_handler(ReqTypes.SUSPEND, self.__suspend_handler)
        self.set_handler(ReqTypes.RESUME, self.__resume_handler)
        self.set_handler('feature', self.__feature_handler)

#
# Main
#

if __name__ == "__main__":
    PalReplication(option=palcmd.PalCmdOption()).handle_request()
