#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for USER configuration."""

from iobase.cmdexec import (CmdExec, NULL, PIPE)
from iobase.exception import (ExecError, InvalidArgError)
import libpal.palcmd as palcmd
from pal_info import (ReqTypes, Results)
import pal_config as config
import libpal.util as util


SNMP_NAME = "snmp"

INVALID_ROOT_TAG_MSG = "Invalid tag name!!"

ENTERPRISE_ID = ".1.3.6.1.4.1.4615"
LANDISK_ID = ENTERPRISE_ID + ".8"
HDLH_ID = LANDISK_ID + ".1"
OBJECT_ID = HDLH_ID + ".1"
INFOMATION_ID = OBJECT_ID + ".1"
MSGCODE_ID = INFOMATION_ID + ".1"
ARGUMENT_ID = INFOMATION_ID + ".2"
NOTIFICATION_ID = HDLH_ID + ".2"
ERROR_NOTIFICATION_ID = NOTIFICATION_ID + ".1"
DISK_SUBTREE_ID = ".1.3.6.1.4.1.2021.9.1"

HOST_PORT_IDX = 6
ERRCODE_IDX = 11
ERRARGS_IDX = 14
OID_IDX = 7

SNMPTRAP_CMD_ARGS = [
    config.SNMPTRAP_CMD,
    "-v", "2c", "-Ci",
    "-c", "public", "", "",
    ERROR_NOTIFICATION_ID, MSGCODE_ID,
    "i", "",
    ARGUMENT_ID, "s", ""]
SNMPGET_CMD_ARGS = [
    config.SNMPGET_CMD,
    "-v", "2c", "-Oqv",
    "-c", "public", "localhost", ""]

HOST_PATH = "trap/host"
PORT_PATH = "trap/port"
CODE_PATH = "messageinfo/code"
ARG_LIST_PATH = "messageinfo/arglist/arg"
OBJECT_NAME = "object"
OID_ATTR = "oid"

SEND_FAILED_MSG = "Fail sending!!"
QUERY_FAILED_MSG_FMT_STR = "Fail querying!!: %s"

SUCCESS_CMD_RETURN_CODE = 0

INTERNAL_MNT_POINT = \
    config.VOL_INFO[config.STOR_IF_TO_VOL["internal"][1]]["mnt_point"]

UNMOUNT_DISK_VALUE = "0"


class PalSnmp(palcmd.PalCmd):
    """PAL Snmp command implementation class."""

    def __check_root_tag(self, entry):
        """Check root tag."""

        root_name = entry.get().get_node_name()

        if root_name != SNMP_NAME:
            raise InvalidArgError(INVALID_ROOT_TAG_MSG)

    def __pipe_wait(self, proc, cmd_args):
        """Pipe wait."""

        proc.wait()
        if proc.returncode != SUCCESS_CMD_RETURN_CODE:
            raise ExecError(cmd_args)

    def __is_disk_oid_and_unmounted(self, oid):
        """Check whether oid is disk concerned value and
           internal disk is umountted.
           If not mounted return True.
        """

        if oid.startswith(DISK_SUBTREE_ID):
            if not util.is_mounted(INTERNAL_MNT_POINT):
                return True

        return False

    def __send_handler(self, entry):
        """Handle send request."""

        try:
            self.__check_root_tag(entry)

            host = entry.get().get_node(HOST_PATH).get_text()
            port = entry.get().get_node(PORT_PATH).get_text()
            errcode = entry.get().get_node(CODE_PATH).get_text()
            errargs = ",".join([
                node.get_text()
                for node
                in entry.get().get_node_list(ARG_LIST_PATH)])

            cmd_args = list(SNMPTRAP_CMD_ARGS)
            cmd_args[HOST_PORT_IDX] = ":".join((host, port))
            cmd_args[ERRCODE_IDX] = errcode
            cmd_args[ERRARGS_IDX] = errargs

            CmdExec(cmd_args).execute(stdout=NULL, stderr=NULL)

            entry.set_result(Results.OK)

        except ExecError:
            entry.set_result(Results.SEND_ERROR)
            self.exception(SEND_FAILED_MSG)

        return entry

    def __query_handler(self, entry):
        """Handle query request."""

        try:
            self.__check_root_tag(entry)

            oid = entry.get().get_node(OBJECT_NAME).get_attr_value(OID_ATTR)

            # ディスク関係のOIDの場合はアンマウント状態の時の特別対応
            if self.__is_disk_oid_and_unmounted(oid):
                entry.get().get_node(OBJECT_NAME).set_text(UNMOUNT_DISK_VALUE)
                entry.set_result(Results.OK)
                return entry

            cmd_args = list(SNMPGET_CMD_ARGS)
            cmd_args[OID_IDX] = oid
            proc = CmdExec(cmd_args).execute(stdout=PIPE, stderr=NULL)
            entry.get().get_node(OBJECT_NAME).set_text(proc.stdout.read())

            self.__pipe_wait(proc, cmd_args)

            entry.set_result(Results.OK)

        except ExecError, ins:
            entry.set_result(Results.QUERY_ERROR)
            self.exception(QUERY_FAILED_MSG_FMT_STR % str(ins.info))

        return entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.SEND, self.__send_handler)
        self.set_handler(ReqTypes.QUERY, self.__query_handler)

#
# Main
#

if __name__ == "__main__":
    PalSnmp(option=palcmd.PalCmdOption()).handle_request()
