# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

import os
import stat
import rapscomm
import time
import hashlib
import fcntl

from PIL import Image
from PIL.ExifTags import TAGS
from stat import ST_ATIME
from operator import itemgetter

import commands
import traceback

class FileLocker:
    path = None
    fp = None

    def __init__(self,path=None):
        self.path = path

    def __del__(self,path=None):
        self.unlock()

    def lock(self):
        if self.fp != None:
            return

        self.fp = open(self.path, "w")
        fcntl.flock(self.fp, fcntl.LOCK_EX)

    def unlock(self):
        if self.fp == None:
            return

        fcntl.flock(self.fp, fcntl.LOCK_UN)
        self.fp.close()
        self.fp = None
        try:
            os.remove(self.path)
        except:
            pass

class ExifInfo:
    '''
    Acquisition of Exif information
    '''
    c_comm = None

    d_exif = None
    
    def __init__(self, comm):
        '''
        Constructor
        '''
        self.c_comm = comm
        self.d_exif = { "ImageDescription"       : "imagedescription",
                        "Make"                   : "make",
                        "Model"                  : "model",
                        "Orientation"            : "orientation",
                        "XResolution"            : "xresolution",
                        "YResolution"            : "yresolution",
                        "ResolutionUnit"         : "resolutionunit",
                        "Software"               : "software",
                        "DateTime"               : "datetime",
                        "Copyright"              : "copyright",
                        "DateTimeOriginal"       : "datetimeoriginal",
                        "CompressedBitsPerPixel" : "compressedbitsperpixel",
                        "ShutterSpeedValue"      : "shutterspeedvalue",
                        "ApertureValue"          : "aperturevalue",
                        "BrightnessValue"        : "brightnessvalue",
                        "ExposureBiasValue"      : "exposurebiasvalue",
                        "MaxApertureValue"       : "maxaperturevalue",
                        "SubjectDistance"        : "subjectdistance",
                        "MeteringMode"           : "meteringmode",
                        "LightSource"            : "lightsource",
                        "Flash"                  : "flash",
                        "ExifImageWidth"         : "exifimagewidth",
                        "ExifImageHeight"        : "exifimageheight" }

    def GetContent(self, path):
        '''
        Acquisition of Exif information
        '''

        # initialize
        s_content = ""

        # suffix type check
        try:
            suffix = os.path.splitext(path)[1].lower()
            if (suffix in rapscomm.suffix_picture) == False:
                return s_content
        except:
            return s_content

        if suffix == ".jpeg" or suffix == ".jpg": 
            s_exifinfo = ""
            try:
                img = Image.open(path)
                exif = img._getexif()
                for id, val in exif.items():
                    tag = TAGS.get(id, id)
                    s_exifinfo += self._getExifInfo(tag, val, 3)
            except:
                return s_content
    
            if len(s_exifinfo) > 0:
                s_content  = "  <exif>\n"
                s_content += s_exifinfo
                s_content += "  </exif>\n"
            
        return s_content

    def _getExifInfo(self, tag, val, space):
        '''
        serch exif
        '''
        s_value = ""
        if tag:
            for s_ekey, s_tag in self.d_exif.items():
                if s_ekey == tag:
                    s_work   = str(val)
                    # modify value to like output of jhead command,
                    # because old implementation was using jhead command.
                    if isinstance(val, tuple):
                        s_work = '/'.join(str(elem) for elem in val)
                    s_value  = " " * space
                    s_value += "<"
                    s_value += s_tag
                    s_value += ">"
                    try:
                        s_value += self.c_comm.Escape(s_work)
                    except:
                        s_value += s_work
                    s_value += "</"
                    s_value += s_tag
                    s_value += ">\n"
                    try:
                        del self.d_exif[s_ekey]
                    except:
                        pass
                    break

        return s_value

class Thumbnail:
    '''import stat

    Thumbnail operation class  
    '''
    c_main = None
    c_comm = None

    CACHE_PATH = ""
    LOCK       = ""

    def __init__(self, main, comm):
        '''
        Constructor
        '''
        self.c_main = main
        self.c_comm = comm
        self.CACHE_PATH = self.c_comm.CONFIG.CACHE_PATH.rstrip("/")
        self.CACHE_PATH += "/"
        self.LOCK_PATH = self.c_comm.CONFIG.LOCK_PATH

    def _remove_old_files(self,path,threshold,buffer_percent=3):
        '''
        remove old files
        '''
        files = os.listdir(path)
        c_count = len(files)
        if c_count < threshold:
            return
        overrun_count = c_count - threshold
        fixed_count = 1
        buffer_count = threshold * buffer_percent / 100
        d_count = overrun_count + fixed_count + buffer_count
        cmd = "ls -rt " + path
        cmd += " | head -n " + str(d_count)
        cmd += " | xargs -n 1 -i{} rm " + path + "/{}"
        ( status, output ) = commands.getstatusoutput(cmd)
        if status != 0:
            raise

    def create(self, srcpath, dstpath="not use this param"):
        '''
        return created thumbnail path. 
        '''

        try:
            flocker = FileLocker(self.LOCK_PATH + "/getmminfo")
            flocker.lock()
        except:
            traceback.print_exc(file=open(self.c_comm.CONFIG.LOG,"w"))
            return ""

        try:
            # initialize
            b_picture = False
            b_movie = False

            # get config size
            width  = self.c_comm.CONFIG.THUMB_WIDTH
            height = self.c_comm.CONFIG.THUMB_HEIGHT

            # create cache name
            h_name = self._create_hash_name( srcpath )

            # create cache path
            s_save = os.path.join( self.CACHE_PATH, h_name )

            # thumbnail cache already exists. 
            if os.path.exists(s_save) == True:
                # update timestamp
                os.utime(s_save, None)
                flocker.unlock()
                return s_save

            # type check
            suffix = os.path.splitext(srcpath)[1].lower()
            if (suffix in rapscomm.suffix_picture) == True:
                b_picture = True
            elif (suffix in rapscomm.suffix_movie) == True:
                b_movie = True
            else:
                flocker.unlock()
                return ""

            self._remove_old_files(self.CACHE_PATH, self.c_comm.CONFIG.THUMB_COUNT)

            if b_movie == True:
                # call fal_command
                if self.c_main != None:
                    if self.c_main._makethumbnail(srcpath, s_save) == self.c_comm.RESULT.RESULT_NO_ERROR:
                        flocker.unlock()
                        return s_save
        
            if b_picture == True:
                img = Image.open(srcpath)

                filter = self.c_comm.CONFIG.THUMB_FILTER
                img.thumbnail( (width, height), getattr(Image, filter))

                if img.mode != "RGB":
                    img = img.convert("RGB")
            
                try:
                    img.save(s_save, "JPEG")
                    os.chmod(s_save, 0666)
                except:
                    self.c_comm.Remove(s_save)
                    flocker.unlock()
                    traceback.print_exc(file=open(self.c_comm.CONFIG.LOG,"w"))
                    return ""

                flocker.unlock()
                return s_save

        except:
            flocker.unlock()
            traceback.print_exc(file=open(self.c_comm.CONFIG.LOG,"w"))
            return ""

    def delete(self, path):
        self.c_comm.logout("called meaningless function")
        return True

    def _create_hash_name(self, path):
        '''
        return cache file name
        '''
        fst = os.stat( path )
        time_fmt = time.strftime("%Y%m%d%H%M%S",time.gmtime(fst.st_mtime))
        size_fmt = str(fst.st_size)

        name = path + "_" + time_fmt + "_" + size_fmt
        h_name = hashlib.md5( name ).hexdigest()

        return h_name

